from typing import Iterable

from rich.progress import BarColumn, Progress, TaskProgressColumn, TextColumn
from tqdm import tqdm


def tqdm_progress(iterable: Iterable, description: str):
    return tqdm(iterable, desc=description)


def rich_progress(iterable: Iterable, description: str, color: str = 'green'):
    total = len(iterable)
    with Progress(
        TextColumn('[progress.description]{task.description}'),
        BarColumn(),
        TaskProgressColumn(),
        TextColumn('{task.completed}/{task.total}'),
    ) as progress:
        task = progress.add_task(f'[{color}]{description}', total=total)
        for i, item in enumerate(iterable, 1):
            yield item
            progress.update(task, advance=1, description=f'[{color}]{description} ({i}/{total})')



from .fabric import BasicFabricTrainer, callbacks, default_fabric, train_epoch
from .fabric.callbacks import (
    FabricCallbackProtocol,
    ParseCSVLogfile,
    PrintTrainInfo,
    SaveCheckpoints,
    SaveHparams,
    find_callback,
)
from .lightning_modules import BasicLightningTrainer
from .metric import TorchMetricRecorder
from .utils import cuda_available, default_device, seed_everything, use_device

__all__ = [
    'cuda_available',
    'default_device',
    'seed_everything',
    'use_device',
    'BasicLightningTrainer',
    'TorchMetricRecorder',
    'BasicFabricTrainer',
    'default_fabric',
    'callbacks',
    'train_epoch',
    'find_callback',
    'FabricCallbackProtocol',
    'ParseCSVLogfile',
    'PrintTrainInfo',
    'SaveCheckpoints',
    'SaveHparams',
]



import math
from collections import OrderedDict
from datetime import datetime
from typing import List, TypedDict

import numpy as np


class EDFHeader(TypedDict):
    version: str
    patient_id: str
    recording_id: str
    start_date: str
    start_time: str
    header_bytes: int
    reserved: str
    n_records: int
    duration: float
    n_signals: int
    label: List[str]
    transducer_type: List[str]
    physical_dimension: List[str]
    physical_min: List[float]
    physical_max: List[float]
    digital_min: List[int]
    digital_max: List[int]
    prefiltering: List[str]
    samples_per_record: List[int]
    ns_reserved: List[str]


class EDF:
    def __init__(self, edf_file_path: str, convert_header=True) -> None:
        self.edf_file_path = edf_file_path
        self._is_edf_plus = False
        self.header = {}
        self._read_edf_header(convert_header)
        self._check_edf_plus()
        self._label_and_srate(self.header)

    def _read_edf_header(self, convert=True):
        with open(self.edf_file_path, 'r+', encoding='iso-8859-1') as f:
            header = OrderedDict(())

            item_names = [
                'version',
                'patient_id',
                'recording_id',
                'start_date',
                'start_time',
                'header_bytes',
                'reserved',
                'n_records',
                'duration',
                'n_signals',
            ]
            item_bytes = [8, 80, 80, 8, 8, 8, 44, 8, 8, 4]
            for name, bytes in zip(item_names, item_bytes):
                header[name] = f.read(bytes)
            ns = int(header['n_signals'])
            item_names = [
                'label',
                'transducer_type',
                'physical_dimension',
                'physical_min',
                'physical_max',
                'digital_min',
                'digital_max',
                'prefiltering',
                'samples_per_record',
                'ns_reserved',
            ]
            item_bytes = [16, 80, 8, 8, 8, 8, 8, 80, 8, 32]
            for name, bytes in zip(item_names, item_bytes):
                header[name] = [f.read(bytes) for _ in range(ns)]

        def convert_type(name):
            if name == 'header_bytes':
                header[name] = int(header[name])
            elif name == 'n_signals':
                header[name] = int(header[name])
            elif name == 'n_records':
                header[name] = int(header[name])
            elif name == 'duration':
                header[name] = float(header[name])
            elif name == 'samples_per_record':
                header[name] = [int(x) for x in header[name]]
            elif name == 'physical_min':
                header[name] = [float(x) for x in header[name]]
            elif name == 'physical_max':
                header[name] = [float(x) for x in header[name]]
            elif name == 'digital_min':
                header[name] = [int(x) for x in header[name]]
            elif name == 'digital_max':
                header[name] = [int(x) for x in header[name]]

        if convert:
            for name, value in header.items():
                if isinstance(value, list):
                    header[name] = [x.strip() for x in value]
                else:
                    header[name] = value.strip()
                convert_type(name)
        self.header = header
        return header

    def _check_edf_plus(self):
        """检查 EDF+ 格式

        - 处理 subsecond
        """
        self.header['subsecond'] = 0
        header = self.header
        if header['label'][-1] == 'EDF Annotations':
            self._is_edf_plus = True
            record_samples = sum(header['samples_per_record'][:-1])
            record_bytes = record_samples * 2
            header_bytes = header['header_bytes']
            with open(self.edf_file_path, 'rb') as f:
                f.seek(header_bytes + record_bytes)  # move to first byte of annotations
                TAL = f.read(header['samples_per_record'][-1] * 2)  # read all annotations
                # print(TAL)
            subsecond_byte = TAL.split(b'\x14\x14')[0]
            subsecond = subsecond_byte.decode('ascii')[
                1:
            ]  # convert to string, and ommit the first character '+'
            # print(offset)
            self.header['subsecond'] = float(subsecond)

    @property
    def start_datetime(self):
        """获取 EDF 文件的开始时间

        Returns
        ----------
            - `datetime.datetime` : 开始时间
        """
        header = self.header
        start_date = header['start_date']
        start_time = header['start_time']
        subsecond = header['subsecond']  # float in seconds
        start_datetime = datetime.strptime(
            f'{start_date} {start_time}', '%d.%m.%y %H.%M.%S'
        )
        # add subsecond to start_datetime
        start_datetime = start_datetime.replace(microsecond=int(subsecond * 1e6))
        return start_datetime

    def _label_and_srate(self, header):
        """获取 Header 中的标签和采样率
        
            自动去除 Annotation 标签，并计算采样率
        
            Parameters
            ----------
                - `header` : dict\\
                    `read_edf_header` 返回的字典
                    
            Returns
            ----------
                - `labels` : ndarray; 标签列表
                - `srate` : ndarray; 采样率s
        """
        self.data_signal_idx = [
            i
            for i in range(len(header['label']))
            if header['label'][i] != 'EDF Annotations'
        ]
        self.label = np.array(header['label'])[self.data_signal_idx]
        self.srate = (
            np.array(header['samples_per_record'])[self.data_signal_idx]
            / header['duration']
        )

    def read_signal(self, start_record=0, end_record=None, channels=None):
        """读取 EDF 文件的信号数据

        Args:
        ----------
            - start_record (int, optional): 从第几个记录开始读取. Defaults to 0.
            - end_record (int, optional): 读取到第几个记录结束. Defaults to None.
            - channels (list, optional): 读取哪些通道. Defaults to None.

        Returns:
        ----------
            - signals: list of np.ndarray, 信号数据
        """
        offset = self.header['header_bytes']
        num_signals = self.header['n_signals']  # C
        num_records = self.header['n_records']  # T
        samples_per_record: list = self.header['samples_per_record']  # N/T

        bytes_per_record = sum(samples_per_record) * 2

        end_record = num_records if end_record is None else end_record
        num_records = end_record - start_record

        # 确定读取的通道
        if channels:
            if 'EDF Annotations' in channels:
                raise ValueError('"EDF Annotations" channel is not supported.')
            data_signal_idx = [self.header['label'].index(x) for x in channels]
        else:
            data_signal_idx = self.data_signal_idx
        # 输入的通道信息的顺序
        signal_idx_order = {x: i for i, x in enumerate(data_signal_idx)}

        signals = [
            np.zeros(samples_per_record[i] * num_records, dtype=np.int16)
            for i in data_signal_idx
        ]

        per_byte = 2
        with open(self.edf_file_path, 'rb+') as f:
            f.seek(offset)
            f.seek(start_record * bytes_per_record, 1)
            for r in range(num_records):
                for s in range(num_signals):
                    spr = samples_per_record[s]
                    if s in data_signal_idx:
                        data_bytes = f.read(spr * per_byte)
                        signal = np.frombuffer(data_bytes, dtype=np.int16)
                        i = signal_idx_order[s]
                        signals[i][r * spr : (r + 1) * spr] = signal
                    else:
                        f.seek(spr * per_byte, 1)

        pmin = np.array(self.header['physical_min'])[data_signal_idx]
        pmax = np.array(self.header['physical_max'])[data_signal_idx]
        dmin = np.array(self.header['digital_min'])[data_signal_idx]
        dmax = np.array(self.header['digital_max'])[data_signal_idx]
        for i in range(len(signals)):
            scale = (pmax[i] - pmin[i]) / (dmax[i] - dmin[i])
            dc = pmax[i] - scale * dmax[i]
            signals[i] = signals[i] * scale + dc

        return signals

    def read_range_signal(self, start_t=0, end_t=None, channels=None):
        if start_t == 0 and end_t is None:
            return self.read_signal(channels=channels)

        duration = self.header['duration']
        samples_per_record = self.header['samples_per_record']
        start_record = int(start_t // duration)
        start_record_offset = start_t % duration
        if math.isclose(end_t % duration, 0):
            # 如果整除的话，ceil 会导致少读取一段，所以需要额外的处理
            end_record = int(end_t // duration) + 1
        else:
            end_record = math.ceil(end_t / duration)
        end_record_offset = duration - end_t % duration
        sr_channels = [
            samples_per_record[self.header['label'].index(x)] for x in channels
        ]

        signals = self.read_signal(start_record, end_record, channels)

        for i in range(len(signals)):
            signal = signals[i]
            sr = sr_channels[i]
            sig_len = len(signal)
            start = int(start_record_offset * sr)
            end = sig_len - int(end_record_offset * sr)
            signals[i] = signal[start:end]
        return signals

    def read_annotation(self):
        """读取 EDF 文件的注释

        Returns:
        ----------
            - annotations: list of dict, 注释数据
        """
        ...


def read_edf_header(fpath: str, convert_header=True):
    """读取 EDF 文件的 Header 信息


    Parameters
    ----------
        - `fpath` : str
            - EDF 文件路径
        - `convert_header` : bool, optional
            - 是否转换 Header 中的数据类型

    Returns
    ----------
        - `OrderedDict` : dict
            - header 对象
    """
    edf = EDF(fpath, convert_header)
    return edf.header



from zyplib.train import fabric

from . import dataset, feature, io, nn, preprocessing, signal, train, utils, vis
from ._config import config  # 初始化

__all__ = [
    'fabric',
    'dataset',
    'feature',
    'io',
    'nn',
    'preprocessing',
    'signal',
    'train',
    'utils',
    'vis',
]



"""
Functional programming tools
"""

import inspect
from functools import lru_cache, wraps
from typing import Any, Callable, List, Protocol

__all__ = ['compose', 'pipe']


class ComposeProtocol(Protocol):
    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...

    def get_intermediate_results(self) -> List[Any]: ...


class PipelineError(Exception):
    """Custom exception for pipeline errors."""


@lru_cache(maxsize=None)
def get_func_signature(func: Callable) -> inspect.Signature:
    """Cache and return the signature of a function."""
    return inspect.signature(func)


def should_unpack(input_value: Any, func: Callable) -> bool:
    """Determine whether the input value should be unpacked for the given function."""
    if not isinstance(input_value, tuple):
        return False

    sig = get_func_signature(func)
    params = list(sig.parameters.values())

    # If there's a *args parameter, we should always unpack
    if any(param.kind == inspect.Parameter.VAR_POSITIONAL for param in params):
        return True

    # Count the number of non-default parameters
    non_default_params = sum(
        1 for param in params if param.default == inspect.Parameter.empty
    )

    # If the number of items in the tuple matches the number of non-default parameters,
    # or if there are more items than parameters (extra will go to *args if present),
    # we should unpack
    return len(input_value) == non_default_params or len(input_value) > len(params)


def compose(*funcs: Callable) -> ComposeProtocol:
    """
    创建一个复合函数，将多个函数串联起来，依次执行。

    Parameters
    ----------
    - `*funcs` : Callable
        - 要串联的函数列表

    Returns
    ----------
    - `ComposeProtocol`
        - 一个管道执行函数
        - 该函数可以像普通函数一样调用，并返回最终结果
        - 该函数还具有 `get_intermediate_results` 方法，可以获取所有中间结果

    Raises
    ----------
    - `PipelineError`
        - _description_

    Example:
    ----------
        >>> def double(x):
        ...     return x * 2
        >>> def add_one(x):
        ...     return x + 1
        >>> pipe = make_pipe(double, add_one)
        >>> result = pipe(3)
        >>> print(result)
        7
        >>> print(pipe.get_intermediate_results())
        [6, 7]
    """

    if not funcs:
        return lambda x: x  # 返回恒等函数

    @wraps(funcs[0])
    def pipeline(*args: Any) -> Any:
        def execute():
            current_value = args
            for i, func in enumerate(funcs):
                try:
                    if i == 0:
                        result = func(*current_value)
                    else:
                        result = (
                            func(*current_value)
                            if should_unpack(current_value, func)
                            else func(current_value)
                        )
                    yield result
                    current_value = result
                except Exception as e:
                    raise PipelineError(
                        f"Error in function '{func.__name__}' at step {i+1}: {str(e)}"
                    ) from e

        pipeline.intermediate_results = list(execute())
        return pipeline.intermediate_results[-1]

    def get_intermediate_results() -> List[Any]:
        return getattr(pipeline, 'intermediate_results', [])

    pipeline.get_intermediate_results = get_intermediate_results
    return pipeline


def pipe(inputs: Any, *funcs: Callable) -> Any:
    """
    创建一个管道函数，将多个函数串联起来，依次执行。

    Parameters
    ----------
    - `inputs` : Any
        - 要传递给第一个函数的输入值
    - `*funcs` : Callable
        - 要串联的函数列表f

    Returns
    ----------
    - `Any`: 执行结果
    """
    if not funcs:
        return inputs

    composed = compose(*funcs)
    return composed(inputs)



import torch
from torch import nn

from zyplib.nn import base


class DepthwiseSeperableConv(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size,
        stride=1,
        padding=0,
        dilation=1,
        bias=False,
        Conv=nn.Conv1d,
    ):
        super(DepthwiseSeperableConv, self).__init__()
        self.depthwise_conv = Conv(
            in_channels,
            in_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            in_channels,
            bias,
        )
        self.pointwise_conv = Conv(in_channels, out_channels, 1, 1, 0, 1, 1, bias)

    def forward(self, x):
        x = self.depthwise_conv(x)
        x = self.pointwise_conv(x)
        return x


class MaxNormConv(nn.Module):
    def __init__(self, conv: base.Conv, max_norm: int = 1):
        super(MaxNormConv, self).__init__()
        self.conv = conv
        self.max_norm = max_norm

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.conv.weight.data = torch.renorm(
            self.conv.weight.data, p=2, dim=0, maxnorm=self.max_norm
        )
        return self.conv.forward(x)



import os
import re
from pathlib import Path

import numpy as np
import scipy.io as sio
from mne.io import read_raw_edf

from zyplib.annotation.event import Event

from ._base import EventMixin


class EDFFile(EventMixin):
    def __init__(self, filepath: str, events: list[Event] | None = None, **mne_kwargs):
        EventMixin.__init__(self, events)
        self.filepath = Path(filepath)
        options = {'verbose': False, 'preload': False}
        options = {**options, **mne_kwargs}
        self.raw = read_raw_edf(filepath, **options)
        self._resolve_raw()

    def _resolve_raw(self):
        # 1. 从 raw 中提取事件
        events = self.raw.annotations
        for event in events:
            self.add_event(Event(event['description'], event['onset'], event['duration']))

    @property
    def meta(self) -> dict:
        return {
            'filepath': self.filepath,
            'n_channels': self.n_channels,
            'srate': self.srate,
            't_duration': self.t_duration,
        }

    @property
    def name(self) -> str:
        return os.path.basename(self.filepath)

    @property
    def channels(self) -> list[str]:
        return self.raw.ch_names

    @property
    def n_channels(self) -> int:
        return len(self.raw.ch_names)

    @property
    def srate(self) -> float:
        return self.raw.info['sfreq']

    @property
    def t_duration(self) -> float:
        return self.raw.times[-1]

    @property
    def n_samples(self) -> int:
        return self.raw.n_times

    @property
    def data(self) -> np.ndarray:
        return self.raw.get_data()

    @property
    def units(self):
        return {ch['ch_name']: ch['unit'] for ch in self.raw.info['chs']}

    def get_epoch(self, t_start: float, t_duration: float, units='uV') -> np.ndarray:
        start_sample = int(t_start * self.srate)
        end_sample = start_sample + int(t_duration * self.srate)
        return self.raw.get_data(start=start_sample, stop=end_sample, units=units)



from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Protocol,
    TypeAlias,
    TypeVar,
    Union,
)

import numpy as np
import torch
from lightning.fabric import Fabric
from lightning.fabric.wrappers import _FabricModule, _FabricOptimizer
from simple_inject import Inject, auto_inject
from torch import nn
from torch.utils.data import DataLoader

from zyplib.train.metric.torch_metric import TorchMetricRecorder
from zyplib.train.utils import step_lr_sched
from zyplib.utils.progress_bar import rich_progress

# 类型定义
Batch: TypeAlias = tuple[torch.Tensor, ...]
Metrics: TypeAlias = Dict[str, float]
T = TypeVar('T')


class MetricRecorder(Protocol):
    def update(
        self, y_pred: torch.Tensor, y: torch.Tensor, loss: float | None = None
    ) -> Metrics: ...
    def compute(self) -> Metrics: ...
    def reset(self) -> None: ...


@dataclass
class FabricTrainingContext:
    """训练上下文对象，用于存储训练相关的状态和配置


    Attributes:
    ----------
    - `model` : `nn.Module`
        - 模型
    - `optimizer` : `torch.optim.Optimizer`
        - 优化器
    - `fabric` : `Fabric | None`, optional
        - 可选项, 如果传入了, 则会调用 fabric 的 `backward` 和 `callback`
    - `train_metric_recorder` : `TorchMetricRecorder | None`, optional
        - 用于记录训练时指标的 recorder 实例，默认为 None
    - `val_metric_recorder` : `TorchMetricRecorder | None`, optional
        - 用于记录验证时指标的 recorder 实例，默认为 None
    """

    model: nn.Module
    optimizer: torch.optim.Optimizer
    fabric: Optional[Fabric] = None
    train_metric_recorder: Optional[MetricRecorder] = None
    val_metric_recorder: Optional[MetricRecorder] = None

    def __post_init__(self):
        if self.fabric is None:
            return

        flag1 = isinstance(self.model, _FabricModule)
        flag2 = isinstance(self.optimizer, _FabricOptimizer)
        if flag1 and flag2:
            return
        elif (not flag1) and (not flag2):
            self.model, self.optimizer = self.fabric.setup(self.model, self.optimizer)
        else:
            raise RuntimeError('FabricTrainingContext 被错误地初始化了')


def _safe_mean(values: list[float]) -> float:
    """安全计算平均值，处理空列表情况"""
    return float(np.mean(values)) if values else 0.0


def training_step(
    ctx: FabricTrainingContext,
    batch: Batch,
    fn_loss: nn.Module,
    batch_idx: Optional[int] = None,
    **kwargs,
) -> Dict[str, Any]:
    """单个训练步骤

    Parameters
    ----------
    - `ctx` : `procedure.FabricTrainingContext`
    - `batch` : `tuple[torch.Tensor, torch.Tensor]`
        - 数据批次
    - `fn_loss` : `nn.Module`
        - 损失函数: `fn_loss(y_pred, y)`
    - `batch_idx`: int

    Returns
    ----------
    - `Dict[str, Any]`: 计算的指标; 至少包含 `loss`
    """
    ctx.model.train()
    x, y = batch

    ctx.optimizer.zero_grad()
    y_pred = ctx.model(x)
    loss: torch.Tensor = fn_loss(y_pred, y)

    if ctx.fabric:
        ctx.fabric.backward(loss)
    else:
        loss.backward()

    ctx.optimizer.step()

    loss_value = loss.item()

    if ctx.train_metric_recorder:
        return ctx.train_metric_recorder.update(y_pred, y, loss_value)
    return {'loss': loss_value}


def validation_step(
    ctx: FabricTrainingContext,
    batch: Batch,
    fn_loss: nn.Module,
    batch_idx: Optional[int] = None,
    **kwargs,
) -> Dict[str, Any]:
    """单个验证步骤

    Parameters
    ----------
    - `ctx` : `procedure.FabricTrainingContext`
    - `batch` : `tuple[torch.Tensor, torch.Tensor]`
        - 数据批次
    - `fn_loss` : `nn.Module`
        - 损失函数: `fn_loss(y_pred, y)`
    - `batch_idx`: int

    Returns
    ----------
    - `Dict[str, Any]`: 计算的指标; 至少包含 `loss`
    """
    ctx.model.eval()
    with torch.no_grad():
        x, y = batch
        y_pred = ctx.model(x)
        loss: torch.Tensor = fn_loss(y_pred, y)

    loss_value = loss.item()
    if ctx.val_metric_recorder:
        return ctx.val_metric_recorder.update(y_pred, y, loss_value)
    return {'loss': loss_value}


@auto_inject()
def train_epoch(
    ctx: FabricTrainingContext,
    train_loader: DataLoader,
    fn_loss: nn.Module,
    epoch_idx: Optional[int] = None,
    fn_train_step: Optional[Callable] = Inject('training_step', namespace='train'),
    **train_step_kwargs: Any,
) -> Dict[str, float]:
    """单个训练轮次

    Parameters
    ----------
    - `ctx` : `procedure.FabricTrainingContext`
    - `train_loader` : `DataLoader`
        - 训练数据加载器
    - `fn_loss` : `nn.Module`
        - 损失函数
    - `epoch_idx` : `int | None`, optional
        - 当前轮次索引, 默认为 None
    - `fn_train_step` : `(...) -> Dict[str, Any]`, optional
        - 可选项, 用于计算单个训练步骤的函数, 默认为 `training_step`

    `fn_train_step`:
    ----------
    - @AutoInject: 从 `namespace=train` 自动获取 `training_step`
    - 默认不提供, 使用 `procedure.training_step`; 如果提供了, 则使用提供的函数
    - 接受以下的关键字参数: `fn(ctx: FabricTrainingContext, batch, fn_loss, batch_idx)`
    - 返回值: `Dict[str, Any]`, 至少包含 `loss`
    Returns
    ----------
    - `Dict[str, float]`: 本 Epoch 的平均指标, 至少包含 `loss`
    """
    if fn_train_step is None:
        fn_train_step = training_step

    if ctx.fabric:
        ctx.fabric.call('on_train_epoch_start', epoch=epoch_idx)

    losses = []
    for idx, batch in enumerate(rich_progress(train_loader, 'Training', color='green')):
        metrics = fn_train_step(
            ctx=ctx, batch=batch, fn_loss=fn_loss, batch_idx=idx, **train_step_kwargs
        )
        if ctx.fabric:
            ctx.fabric.call(
                'on_train_batch_end',
                metrics=metrics,
                batch=batch,
                loss=metrics.get('loss', None),
                batch_idx=idx,
            )
        if ctx.train_metric_recorder is None:
            losses.append(metrics['loss'])

    metrics = (
        ctx.train_metric_recorder.compute()
        if ctx.train_metric_recorder
        else {'loss': _safe_mean(losses)}
    )
    if ctx.fabric:
        ctx.fabric.call('on_train_epoch_end', metrics=metrics, epoch=epoch_idx)
    return metrics


@auto_inject()
def val_epoch(
    ctx: FabricTrainingContext,
    val_loader: DataLoader,
    fn_loss: nn.Module,
    epoch_idx: Optional[int] = None,
    fn_val_step: Optional[Callable] = Inject('validation_step', namespace='train'),
    **val_step_kwargs: Any,
) -> Dict[str, float]:
    """单个验证轮次

    Parameters
    ----------
    - `ctx` : `procedure.FabricTrainingContext`
    - `train_loader` : `DataLoader`
        - 训练数据加载器
    - `fn_loss` : `nn.Module`
        - 损失函数
    - `epoch_idx` : `int | None`, optional
        - 当前轮次索引, 默认为 None
    - `fn_val_step` : `(...) -> Dict[str, Any]`, optional
        - 可选项, 用于计算单个训练步骤的函数, 默认为 `validation_step`

    `fn_val_step`:
    ----------
    - @AutoInject: 从 `namespace=train` 自动获取 `validation_step`
    - 默认不提供, 使用 `procedure.validation_step`; 如果提供了, 则使用提供的函数
    - 接受以下的关键字参数: `fn(ctx: FabricTrainingContext, batch, fn_loss, batch_idx)`
    - 返回值: `Dict[str, Any]`, 至少包含 `loss`

    Returns
    ----------
    - `Dict[str, float]`: 本 Epoch 的平均指标, 至少包含 `loss`
    """
    if fn_val_step is None:
        fn_val_step = validation_step

    if ctx.fabric:
        ctx.fabric.call('on_validation_epoch_start', epoch=epoch_idx)

    losses = []
    for idx, batch in enumerate(rich_progress(val_loader, 'Validating', color='blue')):
        metrics = fn_val_step(
            ctx=ctx, batch=batch, fn_loss=fn_loss, batch_idx=idx, **val_step_kwargs
        )
        if ctx.fabric:
            ctx.fabric.call(
                'on_validation_batch_end',
                metrics=metrics,
                batch=batch,
                loss=metrics.get('loss', None),
                batch_idx=idx,
            )
        if ctx.val_metric_recorder is None:
            losses.append(metrics['loss'])

    metrics = (
        ctx.val_metric_recorder.compute()
        if ctx.val_metric_recorder
        else {'loss': _safe_mean(losses)}
    )
    if ctx.fabric:
        ctx.fabric.call('on_validation_epoch_end', metrics=metrics, epoch=epoch_idx)
    return metrics


def train_loop(
    fabric: Fabric,  # 如果使用 train_loop, 则默认必须用 fabric
    model: nn.Module,
    optimizer: torch.optim.Optimizer,
    train_loader: DataLoader,
    val_loader: Optional[DataLoader],
    fn_loss: nn.Module,
    max_epochs: int,
    train_metric_recorder: TorchMetricRecorder | None = None,
    val_metric_recorder: TorchMetricRecorder | None = None,
    lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler] = None,
    lr_sched_freq: int = 1,
    lr_sched_interval: str = 'epoch',
    return_context: bool = False,
) -> Dict[str, List[float]]:
    """基于 Fabric 的 Train Loop; fabric 为必选参数，如果不想用 fabric 可以自行组合 train_epoch etc.

    依赖注入
    ----------
    可以将 `training_step` 和 `validation_step` 函数注入到 `"train"` namespace 下作；如果不注入，使用默认的训练和验证函数。函数需满足如下签名
    ```py
    # 返回中应当至少包含 {'loss': float}
    fn(ctx: FabricTrainingContext, batch, fn_loss, batch_idx) -> dict[str, float]
    ```

    Parameters
    ----------
    - `fabric` : `Fabric`
        - _description_
    - `optimizer` : `torch.optim.Optimizer`
        - _description_
    - `train_loader` : `DataLoader`
        - _description_
    - `val_loader` : `Optional[DataLoader]`
        - _description_
    - `fn_loss` : `nn.Module`
        - _description_
    - `max_epochs` : `int`
        - _description_
    - `train_metric_recorder` : `TorchMetricRecorder | None`, optional
        - _description_, by default None
    - `val_metric_recorder` : `TorchMetricRecorder | None`, optional
        - _description_, by default None
    - `lr_scheduler` : `Optional[torch.optim.lr_scheduler._LRScheduler]`, optional
        - _description_, by default None
    - `lr_sched_freq` : `int`, optional
        - _description_, by default 1
    - `lr_sched_interval` : `str`, optional
        - _description_, by default 'epoch'

    Returns
    ----------
    - `Dict[str, List[float]]`
        - _description_
    """
    context = FabricTrainingContext(
        model=model,
        optimizer=optimizer,
        fabric=fabric,
        train_metric_recorder=train_metric_recorder,
        val_metric_recorder=val_metric_recorder,
    )

    # model, optimizer = fabric.setup(model, optimizer)
    train_loader = fabric.setup_dataloaders(train_loader)
    if val_loader:
        val_loader = fabric.setup_dataloaders(val_loader)

    fabric.call(
        'on_train_start',
        model=model,
        optimizer=optimizer,
        lr_scheduler=lr_scheduler,
        fabric=fabric,
    )

    for epoch in range(max_epochs):
        fabric.call('on_train_epoch_start', epoch=epoch, max_epochs=max_epochs)

        train_metrics = train_epoch(
            ctx=context, train_loader=train_loader, fn_loss=fn_loss, epoch_idx=epoch
        )

        fabric.call('on_train_epoch_end', metrics=train_metrics, epoch=epoch)

        if val_loader:
            fabric.call('on_validation_epoch_start', epoch=epoch)
            val_metrics = val_epoch(
                ctx=context, val_loader=val_loader, fn_loss=fn_loss, epoch_idx=epoch
            )
            fabric.call('on_validation_epoch_end', metrics=val_metrics, epoch=epoch)
        else:
            val_metrics = None

        if lr_scheduler and lr_sched_interval == 'epoch':
            if (epoch + 1) % lr_sched_freq == 0:
                loss = val_metrics['loss'] if val_metrics else train_metrics['loss']
                step_lr_sched(lr_scheduler, loss)

        fabric.call(
            'on_epoch_end',
            epoch=epoch,
            train_metrics=train_metrics,
            val_metrics=val_metrics,
        )

    fabric.call('on_train_end')

    result = {'train': train_metrics, 'val': val_metrics}
    if return_context:
        result['context'] = context
    return result



import numpy as np

from zyplib.utils.ensure import ensure_npy


def segment_signal(
    signal: np.ndarray,
    fs: int,
    seg_duration: float,
    overlap: float = 0,
    begin: float = 0,
    end: float = None,
):
    """Segment a signal into pieces.

    在 signal最后一个维度(时间维度)上分割,
    如果最后一个段的长度小于 `seg_duration`, 则丢弃最后一个段。

    Parameters
    ----------
    - `signal` : array_like, 1-D or 2-D
        - 待分段的信号。可以是1-D或2-D数组, 类型为 `array_like`。
        - 最后一个维度为时间维度
    - `fs` : int
        - 输入信号的采样频率, 单位为Hz, 类型为 `int`。
    - `seg_duration` : float
        - 每个段的持续时间, 单位为秒, 类型为 `float`。
    - `overlap` : float, 可选
        - 相邻段之间的重叠时间, 单位为秒, 类型为 `float`。
        - 默认值为0.0, 表示无重叠
    - `begin` : float, 可选
        - 信号的起始时间, 单位为秒, 类型为 `float`。
        - 默认值为0.0
    - `end` : float, 可选
        - 信号的结束时间, 单位为秒, 类型为 `float`。
        - 默认值为None, 表示使用信号的最后一个时间点作为结束时间

    Returns
    ----------
    - `segments` : ndarray
        - The segmented signal, `[N, ..., L]`.
    """
    if seg_duration <= 0:
        raise ValueError('seg_duration must be positive.')
    if overlap < 0:
        raise ValueError('overlap must be non-negative.')
    if overlap >= seg_duration:
        raise ValueError('overlap must be less than seg_duration.')

    seg_len = int(seg_duration * fs)
    overlap_len = int(overlap * fs)
    step = seg_len - overlap_len
    if step <= 0:
        raise ValueError('overlap must be less than seg_duration.')

    if end is None:
        end = signal.shape[-1] / fs
    if end <= begin:
        raise ValueError('end must be greater than begin.')

    signal = signal[..., int(begin * fs) : int(end * fs)]

    # 计算分段后的信号长度
    length = signal.shape[-1]
    n_seg = (length - overlap_len) // step
    if n_seg <= 0:
        raise ValueError('signal is too short.')

    signal = ensure_npy(signal)

    # 分段
    segments = np.lib.stride_tricks.as_strided(
        signal,
        shape=(n_seg, *signal.shape[:-1], seg_len),
        strides=(step * signal.strides[-1], *signal.strides),
        writeable=False,
    )
    # 从 view 中拷贝数据, 避免修改原始数据
    segments = segments.copy()
    return segments



"""Key-value 相关工具"""

from functools import singledispatch
from itertools import chain
from pathlib import Path
from typing import Any, OrderedDict

from zyplib.utils.fs import read_json, read_yaml
from zyplib.utils.print import print_error


def collect_key_path(target: dict | str | Path, key: str):
    """获取 一个字典中所有制定 key 的对象，并返回其压缩后的对象 {路径: value}

    >>> dic = {
        'loss': 1,
        'avg': { 'loss': 1.2 },
        'Exp': { 'Round0': { 'loss': 2 }, 'Round1': { 'loss': 1 } }
    }
    >>> collect_key_paths(dic, 'loss')
    {'loss': 1, 'avg.loss': 1.2, 'Exp.Round0.loss': 2, 'Exp.Round1.loss': 1}

    - 输入参数可以为 dict 对象：对对象进行检索
    - 也可以为一个路径：检索路径下所有 yml 和 json 文件内的 key，并返回汇总结果
    """
    if isinstance(target, dict):
        return _collect_key_path_core(target, key)
    elif isinstance(target, (str, Path)):
        return _collect_key_path_files(target, key)
    else:
        raise TypeError(f'Invalid target: {target}')


def _collect_key_path_core(target: dict, key: str) -> dict[str, Any]:
    """处理字典类型的输入"""
    result = {}

    def recurse(sub_dict: dict, parent_key: str = ''):
        # 遍历当前字典的键值对
        for k, v in sub_dict.items():
            # 更新路径信息
            new_key = f'{parent_key}.{k}' if parent_key else k
            # 如果当前键等于目标键，保存值
            if k == key:
                result[new_key] = v
            # 如果值是字典，递归调用
            if isinstance(v, dict):
                recurse(v, new_key)

    # 从根字典开始递归
    recurse(target)
    return result


def _collect_key_path_files(
    target: str | Path, key: str
) -> OrderedDict[str, dict[str, Any]]:
    """处理目录路径类型的输入"""
    path = Path(target)
    result = OrderedDict()

    # 递归查看目录中的所有 YAML/JSON 文件
    all_files = chain(path.rglob('*.yaml'), path.rglob('*.yml'), path.rglob('*.json'))
    for fpath in all_files:
        content: dict = {}
        try:
            if fpath.suffix in ('.yaml', '.yml'):
                content = read_yaml(fpath)
            elif fpath.suffix == '.json':
                content = read_json(fpath)
            else:
                continue

            # 提取key-value对
            key_paths = _collect_key_path_core(content, key)
            if key_paths:  # 只保留非空的结果
                result[str(fpath)] = key_paths

        except Exception as e:
            print_error(f'Error reading {fpath}: {repr(e)}')

    return result


class DottableDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._convert_nested_dicts()

    def _convert_nested_dicts(self):
        for key, value in list(self.items()):
            if isinstance(value, dict) and not isinstance(value, DottableDict):
                super().__setitem__(key, DottableDict(value))

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            raise AttributeError(f"'DottableDict' object has no attribute '{key}'")

    def __setattr__(self, key, value):
        if key.startswith('_'):
            super().__setattr__(key, value)
        else:
            self[key] = value

    def _process_nested_key(self, key):
        """处理可能包含点号的键

        Returns:
            tuple: (is_nested, keys)
                - is_nested: 是否是嵌套的键
                - keys: 键的列表
        """
        if not isinstance(key, str):
            return False, None
        if '.' not in key:
            return False, None
        return True, key.split('.')

    def __getitem__(self, key):
        is_nested, keys = self._process_nested_key(key)
        if not is_nested:
            return super().__getitem__(key)

        d = self
        for k in keys:
            d = d[k]
        return d

    def __setitem__(self, key, value):
        if isinstance(value, dict) and not isinstance(value, DottableDict):
            value = DottableDict(value)

        is_nested, keys = self._process_nested_key(key)
        if not is_nested:
            super().__setitem__(key, value)
            return

        d = self
        for k in keys[:-1]:
            if k not in d:
                d[k] = DottableDict()
            d = d[k]
        d[keys[-1]] = value

    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError:
            raise AttributeError(f"'DottableDict' object has no attribute '{key}'")

    def __delitem__(self, key):
        is_nested, keys = self._process_nested_key(key)
        if not is_nested:
            super().__delitem__(key)
            return

        d = self
        for k in keys[:-1]:
            if k not in d:
                raise KeyError(key)
            d = d[k]
        del d[keys[-1]]

    def __contains__(self, key):
        is_nested, keys = self._process_nested_key(key)
        if not is_nested:
            return super().__contains__(key)

        try:
            d = self
            for k in keys:
                d = d[k]
            return True
        except (KeyError, TypeError):
            return False

    def update(self, other=None, **kwargs):
        if other is not None:
            for k, v in other.items():
                if isinstance(v, dict):
                    if k in self and isinstance(self[k], DottableDict):
                        self[k].update(v)
                    else:
                        self[k] = DottableDict(v)
                else:
                    self[k] = v

        for k, v in kwargs.items():
            if isinstance(v, dict):
                if k in self and isinstance(self[k], DottableDict):
                    self[k].update(v)
                else:
                    self[k] = DottableDict(v)
            else:
                self[k] = v



from typing import List, Optional

import torch.nn as nn

from zyplib.nn.modules import BaseModule, XDim, avgpool, conv


class LeNetPro(BaseModule):
    def __init__(
        self,
        in_ch: int,
        in_samples: int,
        n_class: int,
        xdim: XDim = 1,
        num_layers: int = 2,
        feature_maps: Optional[List[int]] = None,
        kernel_sizes: Optional[List[int]] = None,
        fc_sizes: Optional[List[int]] = None,
    ):
        """基于 LeNet 的改进版，支持任意维度的输入

        默认参数下等价于一个 1D 的 LeNet

        注意 num_layers, feature_maps, kernel_sizes 的长度必须一致

        Parameters
        ----------
        - `in_ch` : `int`
            - 输入通道数
        - `in_samples` : `int`
            - 输入样本数
        - `n_class` : `int`
            - 类别数
        - `num_layers` : `int`, optional
            - 特征提取层数，by default 2
        - `feature_maps` : `list`, optional
            - 特征提取层输出通道数，by default [6, 16]
        - `kernel_sizes` : `list`, optional
            - 特征提取层卷积核大小，by default [5, 5]
        - `fc_sizes` : `list`, optional
            - 全连接层大小，by default [120, 84]
        - `xdim` : `XDim`, optional
            - 卷积网络的维度，默认 1D
        """
        super().__init__()
        Conv = conv(xdim)
        AvgPool = avgpool(xdim)

        # 设置默认值
        if feature_maps is None:
            feature_maps = [6, 16]
        if kernel_sizes is None:
            kernel_sizes = [5, 5]
        if fc_sizes is None:
            fc_sizes = [120, 84]

        assert (
            num_layers == len(feature_maps) == len(kernel_sizes)
        ), 'num_layers, feature_maps, and kernel_sizes must have the same length'

        # 构建特征提取层
        layers = []
        in_channels = in_ch
        feature_size = in_samples
        for i in range(num_layers):
            layers.extend(
                [
                    Conv(in_channels, feature_maps[i], kernel_size=kernel_sizes[i]),
                    nn.ReLU(),
                    AvgPool(kernel_size=2),
                ]
            )
            in_channels = feature_maps[i]
            feature_size = (feature_size - kernel_sizes[i] + 1) // 2  # 更新特征图大小

        self.features = nn.Sequential(*layers)

        # 计算全连接层的输入维度
        self.fc_input_dim = feature_maps[-1] * (feature_size ** xdim)

        # 构建分类器
        classifier_layers = []
        fc_input = self.fc_input_dim
        for fc_size in fc_sizes:
            classifier_layers.extend([nn.Linear(fc_input, fc_size), nn.ReLU()])
            fc_input = fc_size
        classifier_layers.append(nn.Linear(fc_input, n_class))

        self.classifier = nn.Sequential(*classifier_layers)

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x



from .base import (
    BaseModule,
    WrapFunc,
    XDim,
    adaptiveavgpool,
    adaptivemaxpool,
    avgpool,
    batchnorm,
    conv,
    convtranspose,
    dropout,
    maxpool,
)

__all__ = [
    'BaseModule',
    'WrapFunc',
    'conv',
    'avgpool',
    'batchnorm',
    'maxpool',
    'adaptiveavgpool',
    'adaptivemaxpool',
    'convtranspose',
    'XDim',
    'dropout',
]



import csv
import operator
from collections import defaultdict
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Optional, Protocol

import numpy as np
import openpyxl
import simple_inject as sij
import torch
from lightning import Fabric

from zyplib.nn.utils import save_checkpoint
from zyplib.train.utils import prefixed_dict
from zyplib.utils.ensure import ensure_scalar
from zyplib.utils.fs import ensure_dir, write_yaml
from zyplib.utils.print import print_debug, print_info, print_warn
from zyplib.utils.time import TicToc, now


def find_callback(
    fabric: Fabric, callback_type: 'FabricCallbackProtocol'
) -> 'FabricCallbackProtocol':
    """查找 fabric 中的回调函数

    Parameters
    ----------
    - `fabric` : `Fabric`
        - 目标 fabric
    - `callback_type` : `FabricCallbackProtocol`
        - 目标回调函数类型
    
    Returns
    ----------
    - `FabricCallbackProtocol`
        - 目标回调函数
    """
    for callback in fabric._callbacks:
        if isinstance(callback, callback_type):
            return callback
    return None


class FabricCallbackProtocol(Protocol):
    def on_train_start(self, **kwargs): ...

    def on_train_end(self, **kwargs): ...

    def on_train_batch_start(self, batch: Any, batch_idx: int, **kwargs): ...

    def on_train_batch_end(self, batch: Any, batch_idx: int, **kwargs): ...

    def on_train_epoch_start(self, **kwargs): ...

    def on_train_epoch_end(self, **kwargs): ...

    def on_validation_start(self, **kwargs): ...

    def on_validation_end(self, **kwargs): ...

    def on_validation_batch_start(self, batch: Any, batch_idx: int, **kwargs): ...

    def on_validation_batch_end(self, batch: Any, batch_idx: int, **kwargs): ...

    def on_validation_epoch_start(self, **kwargs): ...

    def on_validation_epoch_end(self, **kwargs): ...

    def on_test_start(self, **kwargs): ...

    def on_test_end(self, **kwargs): ...

    def on_test_batch_start(self, batch: Any, batch_idx: int, **kwargs): ...

    def on_test_batch_end(self, batch: Any, batch_idx: int, **kwargs): ...

    def on_exception(self, exception: BaseException, **kwargs): ...

    def on_fit_start(self, **kwargs): ...

    def on_fit_end(self, **kwargs): ...


class EnsureFabricMixin:
    def __init__(self, fabric: Optional[Fabric] = None):
        self.fabric = fabric

    def on_train_start(self, **kwargs):
        if self.fabric is not None:
            return
        fabric = kwargs.get('fabric', None)
        if fabric is None:
            classname = self.__class__.__name__
            raise ValueError(f'{classname} 必须有一个 fabric 属性!')
        self.fabric = fabric


class SaveHparams(FabricCallbackProtocol):
    def __init__(self, dir: str, inject_namespace: str = 'default'):
        """自动保存超参数

        将超参数保存到 `dir`/hparams.yaml 下

        `on_train_start`
        ----------
        - 可以将 `hparams` 作为参数传入
        - 如果未传入，则从 `simple_inject` 中获取 namespace 的字典

        Parameters
        ----------
        - `dir` : `str`
            - 保存目录
        - `inject_namespace` : `str`
            - 从 `simple_inject` 中获取超参数的字典

        Hooks
        -----
        - `on_train_start`
            - `hparams` : `Dict[str, Any]`: 超参数; 如果未传入，则从 `simple_inject` 中获取
        """
        self.path = Path(dir) / 'hparams.yaml'
        self.inject_namespace = inject_namespace

    def on_train_start(self, **kwargs):
        hparams = kwargs['hparams']
        if hparams is None:
            hparams = sij.state(self.inject_namespace)
        if hparams:
            write_yaml(self.path, hparams)


class PrintTrainInfo(FabricCallbackProtocol):
    def __init__(self):
        """打印训练信息

        打印训练信息到控制台

        Hooks
        -----
        - `on_train_start`: 打印训练开始信息, 记录训练开始时间
        - `on_train_end`: 打印训练结束信息, 记录训练结束时间
        - `on_train_epoch_start`: 打印当前 epoch 信息
            - `epoch` : `int`: 当前 epoch 数
            - `max_epochs` : `int`: 最大 epoch 数
        - `on_train_epoch_end`: 打印当前 epoch 训练指标
            - `metrics` : `Dict[str, float]`: 当前 epoch 的训练指标
        - `on_validation_epoch_end`: 打印当前 epoch 验证指标
            - `metrics` : `Dict[str, float]`: 当前 epoch 的验证指标
        - `on_test_end`: 打印测试指标
            - `metrics` : `Dict[str, float]`: 测试指标
        """
        self.tictoc = TicToc()

    def _print_epoch_summary(self, metrics: Dict[str, float]):
        for key, value in metrics.items():
            print_info(f'\t{key}: {value:.4f}', end='')
        print()

    def on_train_start(self, **kwargs):
        self.tictoc.tic()
        print_info(f'开始训练: {now()}'.center(80, '='))
        print()

    def on_train_end(self, **kwargs):
        self.tictoc.toc()
        print()
        self.tictoc.print_elapsed('This training')
        print_info(f'结束训练: {now()}'.center(80, '='))

    def on_train_epoch_start(self, epoch: int, max_epochs: int, **kwargs):
        print_info(f'Epoch {epoch}/{max_epochs}'.center(50, '-'))

    def on_train_epoch_end(self, metrics: Dict[str, float], **kwargs):
        print_info('Train |', end='')
        self._print_epoch_summary(metrics)

    def on_validation_epoch_end(self, metrics: Dict[str, float], **kwargs):
        print_info('Val | ', end='')
        self._print_epoch_summary(metrics)

    def on_test_end(self, metrics: Dict[str, float], **kwargs):
        print_info('Test | ', end='')
        self._print_epoch_summary(metrics)


class SaveCheckpoints(FabricCallbackProtocol):
    def __init__(
        self,
        dir: str,
        save_every_k: Optional[int] = None,
        save_best: bool = True,
        keep_best_k: Optional[int] = 1,
        monitor: str = 'loss',
        monitor_on: Literal['train', 'val'] = 'val',
        less_is_better: bool = True,
        ckpt_ext: str = '.ckpt',
    ):
        """自动保存模型检查点

        Parameters
        ----------
        - `dir` : `str`
            - 保存检查点的目录
        - `save_every_k` : `Optional[int]`, optional
            - 每多少个epoch保存一次，默认为None
        - `save_best` : `bool`, optional
            - 是否保存最佳模型，默认为True
        - `keep_best_k` : `Optional[int]`, optional
            - 只保存最佳模型的数量，默认为 1;
        - `monitor` : `str`, optional
            - 用于确定最佳模型的指标，默认为'loss'
            - 可以有 `train_` 和 `val_` 前缀
        - `monitor_on` : `Literal['train', 'val']`, optional
            - 在哪个阶段进行监控，默认为'val', 表示监听 validation 的指标
        - `less_is_better` : `bool`, optional
            - 是否为越小越优，默认为True
        - `ckpt_ext` : `str`, optional
            - 检查点文件的扩展名，默认为'.ckpt'

        Hooks
        -----
        - `on_train_start`
            - `model` : `nn.Module`: 模型
            - `optimizer` : `Optimizer`: 优化器
            - `lr_scheduler` : `_LRScheduler`: 学习率调度器
        - `on_train_epoch_start`: 清空内置的 train 和 val 指标缓存
        - `on_train_epoch_end`
            - `metrics` : `Dict[str, float]`: 更新当前 epoch 的训练指标
        - `on_validation_epoch_end`
            - `metrics` : `Dict[str, float]`: 更新当前 epoch 的验证指标
        - `on_epoch_end`: 保存检查点
        """

        ensure_dir(dir)

        self.dir = Path(dir)
        self.save_every_k = save_every_k
        self.save_best = save_best
        self.keep_best_k = keep_best_k
        self.monitor = monitor
        self.monitor_on = monitor_on
        self.best_score = float('inf') if less_is_better else float('-inf')
        self.current_epoch = 0

        self.score_cmp = operator.lt if less_is_better else operator.gt
        self.ckpt_ext = ckpt_ext

    def on_train_start(self, *, model, optimizer, lr_scheduler, **kwargs):
        self.model = model
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

    def on_train_epoch_start(self, **kwargs):
        self.train_metrics = None
        self.val_metrics = None

    def on_train_epoch_end(self, metrics: dict, **kwargs):
        self.train_metrics = metrics

    def on_validation_epoch_end(self, metrics: dict, **kwargs):
        self.val_metrics = metrics

    def on_epoch_end(self, **kwargs):
        self.current_epoch += 1

        # 定期保存检查点
        if self.save_every_k and self.current_epoch % self.save_every_k == 0:
            checkpoint_path = (
                self.dir / f'checkpoint_epoch={self.current_epoch}{self.ckpt_ext}'
            )
            self._save(checkpoint_path)
            print_debug(f'Saved checkpoint at epoch {self.current_epoch}')

        if not self.save_best:
            return

        metrics = self.val_metrics if self.monitor_on == 'val' else self.train_metrics
        if not metrics:
            return

        current_score = (
            metrics.get(self.monitor)
            or metrics.get(f'val_{self.monitor}')
            or metrics.get(f'train_{self.monitor}')
        )

        if current_score is None:
            return

        if not self.score_cmp(current_score, self.best_score):
            return

        self.best_score = current_score
        best_model_name = self.__format_best_model_fname(
            epoch=self.current_epoch, monitor=self.monitor, score=current_score
        )
        checkpoint_path = self.dir / best_model_name
        self._save(checkpoint_path)
        print_debug(
            f'Saved best model at epoch {self.current_epoch} with {self.monitor}: {current_score:.4f}'
        )

        # 保留最佳 K 个模型
        if self.keep_best_k:
            checkpoints = sorted(
                self.dir.glob(f'best_model_*{self.ckpt_ext}'),
                key=lambda x: self.__extract_best_model_info(x.name)['score'],
                reverse=not self.score_cmp(0, 1),  # True if higher is better
            )
            if len(checkpoints) > self.keep_best_k:
                for checkpoint in checkpoints[self.keep_best_k :]:
                    checkpoint.unlink()

    def __format_best_model_fname(
        self,
        epoch: Optional[int] = None,
        monitor: Optional[str] = None,
        score: Optional[float] = None,
    ):
        epoch_part = f'_epoch={epoch}' if epoch else ''
        monitor_part = f'_on={monitor}' if monitor else ''
        score_part = f'_score={score:.4f}' if score else ''

        return f'best_model{epoch_part}{monitor_part}{score_part}{self.ckpt_ext}'

    def __extract_best_model_info(self, fname: str):
        fname = fname.removesuffix(self.ckpt_ext)
        each_parts = fname.split('_')
        info = {'epoch': None, 'monitor': None, 'score': None}
        for each_part in each_parts:
            if each_part.startswith('epoch='):
                info['epoch'] = int(each_part.split('=')[-1])
            elif each_part.startswith('on='):
                info['monitor'] = each_part.split('=')[-1]
            elif each_part.startswith('score='):
                info['score'] = float(each_part.split('=')[-1])
        return info

    def _save(self, checkpoint_path: str, **kwargs):
        save_checkpoint(
            models=self.model,
            path=checkpoint_path,
            optimizers=self.optimizer,
            epoch=self.current_epoch,
            train_metrics=self.train_metrics,
            val_metrics=self.val_metrics,
            **kwargs,
        )

    def best_checkpoint(self) -> Path:
        dir, ext = self.dir, self.ckpt_ext
        checkpoints = list(Path(dir).glob(f'best_model*{ext}'))
        if not checkpoints:
            raise FileNotFoundError(f'No checkpoints found in {dir} with extension {ext}')

        def extract_score(checkpoint: Path) -> float:
            info = self.__extract_best_model_info(checkpoint.name)
            return info.get(
                'score', float('inf') if self.score_cmp == operator.lt else float('-inf')
            )

        best_checkpoint = max(
            checkpoints,
            key=lambda x: extract_score(x)
            if self.score_cmp == operator.gt
            else -extract_score(x),
        )
        return best_checkpoint  # This is a Path object


class MonitorLearningRate(EnsureFabricMixin, FabricCallbackProtocol):
    def __init__(self, fabric: Optional[Fabric] = None):
        """监控学习率

        监控学习率的变化，并记录到 fabric 的 loggers 中

        Parameters
        ----------
        - `fabric` : `Fabric`
            - 训练使用的 fabric 对象

        Hooks
        -----
        - `on_train_start`
            - `optimizer` : `Optimizer`: 优化器
            - `lr_scheduler` : `_LRScheduler`: 学习率调度器
        - `on_train_epoch_start`
            - `epoch` : `int`: 当前 epoch 数
        """
        EnsureFabricMixin.__init__(self, fabric)

    def on_train_start(self, *, optimizer, lr_scheduler, **kwargs):
        EnsureFabricMixin.on_train_start(self, **kwargs)
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler

    def on_train_epoch_start(self, **kwargs):
        lrs = [group['lr'] for group in self.optimizer.param_groups]
        epoch = kwargs.get('epoch', None)
        for i, lr in enumerate(lrs):
            self.fabric.log(f'lr_group_{i}', lr, epoch)
        if len(lrs) == 1:
            self.fabric.log('learning_rate', lrs[0], epoch)


class FabricLogMetrics(EnsureFabricMixin, FabricCallbackProtocol):
    def __init__(self, fabric: Fabric | None = None):
        """使用 Fabric 记录训练过程中的指标 (`fabric.log_dict`)

        Parameters
        ----------
        - `fabric` : `Fabric`, optional
            - 用于记录指标的 fabric 实例，默认为 None
            - 如果未传入，则会在 `on_train_start` 中从 kwargs 中获取 `fabric` 对象
            - *如果仍然为 None，则会抛出 ValueError

        Hooks
        -----
        - `on_train_start`
            - `fabric` : 可选的 `Fabric`
        - `on_train_batch_end`
            - `metrics` : `Dict[str, Any]`
        - `on_validation_batch_end`
            - `metrics` : `Dict[str, Any]`
        - `on_train_epoch_end`
            - `metrics` : `Dict[str, Any]`
            - `epoch` : `int`
        - `on_validation_epoch_end`
            - `metrics` : `Dict[str, Any]`
            - `epoch` : `int`
        - `on_test_end`
            - `metrics` : `Dict[str, Any]`
        """
        EnsureFabricMixin.__init__(self, fabric)
        self.train_batch_idx = 0
        self.val_batch_idx = 0

    @staticmethod
    def scalar_dict(metrics: Dict[str, Any]) -> Dict[str, Any]:
        """确保 metrics 中的所有值都是标量; 如果无法转换为标量，则直接丢弃"""
        scalar_metrics = {}
        for key, value in metrics.items():
            value = ensure_scalar(value)
            if value is not None:
                scalar_metrics[key] = value
        return scalar_metrics

    def on_train_batch_end(
        self, metrics: Dict[str, Any], *args, step: Optional[int] = None, **kwargs
    ):
        metrics = prefixed_dict('train_', metrics)
        metrics = self.scalar_dict(metrics)
        if step is None:
            step = self.train_batch_idx
            self.train_batch_idx += 1
        else:
            self.train_batch_idx = step
        self.fabric.log_dict(metrics, step=step)

    def on_validation_batch_end(
        self, metrics: Dict[str, Any], *args, step: Optional[int] = None, **kwargs
    ):
        metrics = prefixed_dict('val_', metrics)
        metrics = self.scalar_dict(metrics)
        if step is None:
            step = self.val_batch_idx
            self.val_batch_idx += 1
        else:
            self.val_batch_idx = step
        self.fabric.log_dict(metrics, step=step)

    def on_train_epoch_end(self, metrics: Dict[str, Any], epoch: int, **kwargs):
        metrics = prefixed_dict('train_epoch_', metrics)
        metrics = self.scalar_dict(metrics)
        self.fabric.log_dict(metrics, step=epoch)

    def on_validation_epoch_end(self, metrics: Dict[str, Any], epoch: int, **kwargs):
        metrics = prefixed_dict('val_epoch_', metrics)
        metrics = self.scalar_dict(metrics)
        self.fabric.log_dict(metrics, step=epoch)

    def on_test_end(self, metrics: Dict[str, Any], epoch: Optional[int] = None, **kwargs):
        metrics = prefixed_dict('test_', metrics)
        metrics = self.scalar_dict(metrics)
        self.fabric.log_dict(metrics, step=epoch)


class ParseCSVLogfile:
    def __init__(self, csv_fpath: str, to_excel: bool = True, to_yaml: bool = True):
        self.csv_fpath = Path(csv_fpath)
        self.to_excel = to_excel
        self.to_yaml = to_yaml
        self.excel_fpath = self.csv_fpath.parent / f'{self.csv_fpath.name}.xlsx'
        self.yaml_fpath = self.csv_fpath.parent / f'{self.csv_fpath.name}.yaml'

        if not (to_excel or to_yaml):
            raise ValueError("At least one of 'to_excel' or 'to_yaml' must be True")

    def _process_csv(self):
        # Read CSV file
        with open(self.csv_fpath, 'r', encoding='utf-8') as csvfile:
            reader = csv.DictReader(csvfile)
            data = list(reader)

        # Process and organize data
        organized_data = defaultdict(lambda: defaultdict(list))
        metrics = set()

        for row in data:
            step = int(row.get('step', 0))
            for key, value in row.items():
                if key not in ['step'] and value:
                    metrics.add(key)
                    organized_data[key][step].append(float(value))

        # 不知道为啥 logger 会重复记录，所以这里简单取平均
        repeated_metrics = set()
        for metric in organized_data:
            for step in organized_data[metric]:
                if len(organized_data[metric][step]) > 1:
                    repeated_metrics.add(metric)
                organized_data[metric][step] = sum(organized_data[metric][step]) / len(
                    organized_data[metric][step]
                )
        if repeated_metrics:
            print_warn(f'发现 CSV Logger 存在重复记录的指标: {repeated_metrics}')

        return organized_data, metrics

    def _to_yaml(self, organized_data, metrics):
        names = sorted(list(metrics))
        final_data = {name: dict(organized_data[name]) for name in names}
        write_yaml(self.yaml_fpath, final_data)

    def _to_excel(self, organized_data, metrics):
        wb = openpyxl.Workbook()

        # Create sheets for each metric
        for metric in metrics:
            sheet = wb.create_sheet(title=metric)
            sheet['A1'] = 'Step'
            sheet['B1'] = 'Value'
            row = 2
            for step, value in sorted(organized_data[metric].items()):
                sheet[f'A{row}'] = step
                sheet[f'B{row}'] = value
                row += 1

        # Remove the default sheet created by openpyxl
        wb.remove(wb['Sheet'])

        # Save the Excel file
        wb.save(self.excel_fpath)

    def _convert(self):
        if not self.to_excel and not self.to_yaml:
            return
        if not self.csv_fpath.exists():
            print_warn(f'CSV 日志文件不存在, 无法转换: {self.csv_fpath}')
            return
        organized_data, metrics = self._process_csv()
        if self.to_yaml:
            self._to_yaml(organized_data, metrics)
        if self.to_excel:
            self._to_excel(organized_data, metrics)

    def on_train_end(self, **kwargs):
        self._convert()

    def on_validation_end(self, **kwargs):
        self._convert()

    def on_test_end(self, **kwargs):
        self._convert()


class EarlyStopping(FabricCallbackProtocol):
    def __init__(
        self,
        monitor: str = 'val_loss',
        min_delta: float = 0.0,
        patience: int = 10,
        mode: Literal['min', 'max'] = 'min',
        verbose: bool = True,
        stop_flag: str = 'early_stop',
        fn_stop_callback: Callable[[int], None] | None = None,
    ):
        """Early stopping

        仅仅用于计算何时需要早停，并通过 flag 或者回调函数通知 trainer 停止训练；
        具体的停止训练逻辑需要用户/Traienr自己实现

        使用要求:
        - trainer 和 fn_stop_callback 两者至少要有一个
        - trainer
            1. 在 `on_fit_start` 中传入一个 `trainer` 对象
            2. 通过设置 `trainer` 的 `stop_flag` 属性来通知 tranier 是否停止训练
        - fn_stop_callback: 通过 `fn_stop_callback(epoch)` 来通知外部停止训练


        Parameters
        ----------
        - `monitor` : `str`, 可选
            - 监控的指标，默认为 'val_loss'
        - `min_delta` : `float`, 可选
            - 监控数量变化的最小值，以此作为改进的标准，默认为 0.0
        - `patience` : `int`, 可选
            - 在没有改进的情况下，训练将停止的 epoch 数，默认为 10
        - `mode` : `Literal['min', 'max']`, 可选
            - 在 'min' 模式下，当监控数量不再减少时停止训练；
            - 在 'max' 模式下，当监控数量不再增加时停止训练，默认为 'min'
        - `verbose` : `bool`, 可选
            - 如果为 True，则每次更新时打印消息，默认为 True
        - `stop_flag` : `str`, 可选
            - 设置在 trainer 上的标志以中断训练
        - `fn_stop_callback` : `Callable[[int], None]`, 可选
            - 当训练停止时调用的回调函数，默认为 None

        Hooks
        -----
        - `on_fit_start`
        - `on_validation_epoch_end`
            - `metrics` : `Dict[str, float]`
            - `epoch` : `int`
        """
        self.monitor = monitor
        self.min_delta = np.abs(min_delta)
        self.patience = patience
        self.mode = mode
        self.verbose = verbose
        self.stop_flag = stop_flag
        self.fn_stop_callback = fn_stop_callback

        self.counter = 0
        self.best_score: Optional[float] = None
        self.best_epoch: Optional[int] = None
        self.stopped_epoch = 0
        self.stop_training = False
        self.trainer = None

        assert self.mode in ('min', 'max'), f"EarlyStopping mode '{self.mode}' is unknown"

    def _is_improved(self, current_score: float, best_score: float) -> bool:
        if self.mode == 'min':
            return current_score < best_score - self.min_delta
        else:
            return current_score > best_score + self.min_delta

    def _reset(self) -> None:
        """重置 EarlyStopping 状态"""
        self.best_score = None
        self.best_epoch = None
        self.counter = 0
        self.stopped_epoch = 0
        self.stop_training = False

    def on_fit_start(self, **kwargs) -> None:
        """在训练开始时调用"""
        self._reset()
        self.trainer = kwargs.get('trainer', None)
        if self.trainer is None and self.fn_stop_callback is None:
            print_warn(
                '无法获取到 trainer 对象，且没有设置 fn_stop_callback。EarlyStopping 将不会生效!'
            )

    def on_validation_epoch_end(
        self, metrics: Dict[str, float], epoch: int, **kwargs
    ) -> None:
        """在每个验证 epoch 结束时调用"""
        if self.trainer is None and self.fn_stop_callback is None:
            return

        current_score = metrics.get(self.monitor)
        if current_score is None:
            print_warn(
                f"无法获取到早停条件指标 `{self.monitor}`。"
                f"可用的指标有: {','.join(metrics.keys())}"
            )
            return

        if self.best_score is None:
            self.best_score = current_score
            self.best_epoch = epoch
        elif self._is_improved(current_score, self.best_score):
            self.best_score = current_score
            self.best_epoch = epoch
            self.counter = 0
        else:
            self.counter += 1
            if self.verbose:
                print_debug(f'EarlyStopping counter: {self.counter}/{self.patience}')
            if self.counter >= self.patience:
                self.stopped_epoch = epoch
                self.stop_training = True
                if self.trainer:
                    setattr(self.trainer, self.stop_flag, True)
                if self.verbose:
                    print_info(f'Epoch {epoch}: early stopping')
                if self.fn_stop_callback:
                    self.fn_stop_callback(epoch)



from .cache import (
    cache_factory,
    get_cache,
    make_func_memoize,
    set_cache,
    update_cache_config,
    use_cache,
)
from .cli import parse_cli_args
from .ensure import ensure_2dims, ensure_3dims, ensure_npy, ensure_scalar, ensure_tensor
from .fs import (
    ensure_dir,
    file_reader,
    file_writer,
    read_file,
    read_json,
    read_yaml,
    write_file,
    write_json,
    write_yaml,
)
from .kv import DottableDict, collect_key_path
from .misc import RuntimeDir, random_string
from .print import (
    print_colored,
    print_debug,
    print_error,
    print_info,
    print_warn,
    print_warning,
    type_summary,
)
from .progress_bar import rich_progress, tqdm_progress
from .time import TIME_FMT, TicToc, format_time, format_time_delta, now, print_elapsed

__all__ = [
    'cache_factory',
    'get_cache',
    'make_func_memoize',
    'set_cache',
    'update_cache_config',
    'use_cache',
    'parse_cli_args',
    'ensure_2dims',
    'ensure_3dims',
    'ensure_npy',
    'ensure_scalar',
    'ensure_tensor',
    'ensure_dir',
    'file_reader',
    'file_writer',
    'read_file',
    'read_json',
    'read_yaml',
    'write_file',
    'write_json',
    'write_yaml',
    'random_string',
    'TIME_FMT',
    'TicToc',
    'format_time',
    'format_time_delta',
    'now',
    'print_elapsed',
    'rich_progress',
    'tqdm_progress',
    'print_colored',
    'print_debug',
    'print_error',
    'print_info',
    'print_warn',
    'print_warning',
    'type_summary',
    'DottableDict',
    'RuntimeDir',
    'DottableDict',
    'collect_key_path'
]



import torch
import torch.nn as nn

from zyplib.nn import base
from zyplib.nn.modules.conv import MaxNormConv

# class Conv2dWithConstraint(nn.Conv2d):
#     def __init__(self, *args, max_norm: int = 1, **kwargs):
#         self.max_norm = max_norm
#         super(Conv2dWithConstraint, self).__init__(*args, **kwargs)

#     def forward(self, x: torch.Tensor) -> torch.Tensor:
#         self.weight.data = torch.renorm(
#             self.weight.data, p=2, dim=0, maxnorm=self.max_norm
#         )
#         return super(Conv2dWithConstraint, self).forward(x)


class EEGNet(base.BaseModule):
    def __init__(
        self,
        samples: int = 128,
        channels: int = 60,
        F1: int = 8,
        F2: int = 16,
        D: int = 2,
        n_class: int = 1,
        kernel_1: int = 64,
        kernel_2: int = 16,
        dropout: float = 0.25,
    ):
        """EEGNet v4 版模型; 来源自 torcheeg 的实现

        - Paper: Lawhern V J, Solon A J, Waytowich N R, et al. EEGNet: a compact convolutional neural network for EEG-based brain-computer interfaces[J]. Journal of neural engineering, 2018, 15(5): 056013.
        - URL: https://arxiv.org/abs/1611.08024
        - Related Project: https://github.com/braindecode/braindecode/tree/master/braindecode


        Parameters
        ----------
        - `samples` : `int`, optional
            - 每个样本的采样点数，默认 128
        - `channels` : `int`, optional
            - 每个样本的通道数，默认 60
        - `F1` : `int`, optional
            - 第一个卷积层的卷积核数，默认 8
        - `F2` : `int`, optional
            - 第二个卷积层的卷积核数，默认 16
        - `D` : `int`, optional
            - 深度乘数，默认 2
        - `n_class` : `int`, optional
            - 分类数，默认 1
        - `kernel_1` : `int`, optional
            - 第一个卷积层的卷积核数，默认 64
        - `kernel_2` : `int`, optional
            - 第二个卷积层的卷积核数，默认 16
        - `dropout` : `float`, optional
            - 丢弃概率，默认 0.25
        """

        super(EEGNet, self).__init__()
        self.F1 = F1
        self.F2 = F2
        self.D = D
        self.n_samples = samples
        self.num_classes = n_class
        self.num_electrodes = channels
        self.kernel_1 = kernel_1
        self.kernel_2 = kernel_2
        self.dropout = dropout

        self.block1 = nn.Sequential(
            nn.Conv2d(
                1,
                self.F1,
                (1, self.kernel_1),
                stride=1,
                padding=(0, self.kernel_1 // 2),
                bias=False,
            ),
            nn.BatchNorm2d(self.F1, momentum=0.01, affine=True, eps=1e-3),
            MaxNormConv(
                nn.Conv2d(
                    self.F1,
                    self.F1 * self.D,
                    (self.num_electrodes, 1),
                    stride=1,
                    padding=(0, 0),
                    groups=self.F1,
                    bias=False,
                ),
                max_norm=1,
            ),
            nn.BatchNorm2d(self.F1 * self.D, momentum=0.01, affine=True, eps=1e-3),
            nn.ELU(),
            nn.AvgPool2d((1, 4), stride=4),
            nn.Dropout(p=dropout),
        )

        self.block2 = nn.Sequential(
            nn.Conv2d(
                self.F1 * self.D,
                self.F1 * self.D,
                (1, self.kernel_2),
                stride=1,
                padding=(0, self.kernel_2 // 2),
                bias=False,
                groups=self.F1 * self.D,
            ),
            nn.Conv2d(
                self.F1 * self.D,
                self.F2,
                1,
                padding=(0, 0),
                groups=1,
                bias=False,
                stride=1,
            ),
            nn.BatchNorm2d(self.F2, momentum=0.01, affine=True, eps=1e-3),
            nn.ELU(),
            nn.AvgPool2d((1, 8), stride=8),
            nn.Dropout(p=dropout),
        )

        self.lin = nn.Linear(self.feature_dim(), n_class, bias=False)

    def feature_dim(self):
        with torch.no_grad():
            mock_eeg = torch.zeros(1, 1, self.num_electrodes, self.n_samples)

            mock_eeg = self.block1(mock_eeg)
            mock_eeg = self.block2(mock_eeg)

        return self.F2 * mock_eeg.shape[3]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Input (trials, 1, channels, samples)."""
        if x.ndim == 3:
            x = x.unsqueeze(1)
        x = self.block1(x)
        x = self.block2(x)
        x = x.flatten(start_dim=1)
        x = self.lin(x)

        return x



from typing import Union

import numpy as np
import torch


def to_onehot(
    y: Union[np.ndarray, torch.Tensor, list[int]], squeeze_2_class: bool = True
) -> Union[np.ndarray, torch.Tensor]:
    """将标签编码为独热码

    Parameters
    ----------
    - `y` : `np.ndarray[int] | torch.Tensor[int] | list[int]`
        - 标签, 一个一维的整数数组或张量
    - `squeeze_2_class` : `bool`, optional
        - 如果为 True 且标签为二分类, 则返回一个 [N, 1] 维度的独热码

    Returns
    ----------
    - `np.ndarray` or `torch.Tensor`
        - 独热码

    Raises
    ----------
    - `TypeError`
        - 输入必须是 NumPy 数组或 PyTorch Tensor
    """
    if isinstance(y, np.ndarray):
        num_classes = len(np.unique(y))
        if num_classes == 1 or (num_classes == 2 and squeeze_2_class):
            return y.reshape(-1, 1)
        return np.eye(num_classes)[y]
    elif isinstance(y, torch.Tensor):
        num_classes = len(torch.unique(y))
        if num_classes == 1 or (num_classes == 2 and squeeze_2_class):
            return y.reshape(-1, 1)
        return torch.eye(num_classes, device=y.device)[y]
    elif isinstance(y, list):
        return to_onehot(np.array(y), squeeze_2_class)
    else:
        raise TypeError('Input must be a NumPy array or PyTorch Tensor')


def from_onehot(y: Union[np.ndarray, torch.Tensor]) -> Union[np.ndarray, torch.Tensor]:
    """将独热码转换回标签

    Parameters
    ----------
    - `y` : `np.ndarray` or `torch.Tensor`
        - 独热码, 一个一维或二维的浮点数数组或张量

    Returns
    ----------
    - `np.ndarray` or `torch.Tensor`
        - 标签, 一个一维的整数数组或张量

    Raises
    ----------
    - `TypeError`
        - 输入必须是 NumPy 数组或 PyTorch Tensor
    """
    if isinstance(y, np.ndarray):
        if y.ndim == 1:
            return y.astype(int)
        return np.argmax(y, axis=1)
    elif isinstance(y, torch.Tensor):
        if y.dim() == 1:
            return y.long()
        return torch.argmax(y, dim=1)
    else:
        raise TypeError('Input must be a NumPy array or PyTorch Tensor')



from collections import OrderedDict
from typing import Dict, Optional, Union

import numpy as np
import torch
from torch import nn


def module_grad_norm(module: nn.Module, p_norm: int = 2) -> float:
    """计算 module 当前所有参数的梯度的范数

    moduleGradNorm 的计算结果为 module 中所有子模块的 parameter 组成的向量的 norm

    Parameters:
    ----------
        - `module`: `torch.nn.Module`
        - `p_norm`: `int`, 范数阶数

    Returns:
    ----------
        - `total_grad_norm`: `float`
    """
    params = filter(lambda p: p.grad is not None, module.parameters())
    sum_grad_norm = 0
    for p in params:
        grad = p.grad.detach().norm(p_norm).item()
        sum_grad_norm += grad**p_norm
    total_grad_norm = sum_grad_norm ** (1 / p_norm)
    return total_grad_norm


def param_grad_norm(module: nn.Module, p_norm: int = 2) -> Dict[str, float]:
    """分别计算 module 内各个 parameter 梯度的范数，以字典的格式返回

    Parameters:
    ----------
        - `module`: `torch.nn.Module`
        - `p_norm`: `int`, 范数阶数

    Returns:
    ----------
        - `grads`: `dict`, `{parameter_name: norm}`
    """
    norms = {}
    for name, p in module.named_parameters():
        if p.grad is not None:
            grad = p.grad.detach().norm(p_norm).item()
            norms[name] = grad
    return norms


def grad_norm(
    module: nn.Module, p_norm: int = 2, type: str = 'module'
) -> Union[float, dict]:
    """计算 module 梯度的范数

    根据 type 的不同，采用不同的方式计算; 本质上是调用其他两个函数

    Parameters:
    ----------
        - `module`: `torch.nn.Module`
        - `p_norm`: `int`, 范数阶数
        - `type`:   `int`, 返回类型
            - 'module': 计算整个 module 梯度的范数，调用 `moduleGradNorm`，返回 float
            - 'param':  计算 module 各个 parameter 的范数，调用 `paramGradNorm`，返回 dict

    Returns:
    ----------
        - ans: `Union[float, dict]`
    """
    if type == 'module':
        return module_grad_norm(module, p_norm)
    elif type == 'param':
        return param_grad_norm(module, p_norm)
    else:
        return ValueError(f'Invalid argument type: {type}')



from typing import Literal

import numpy as np

from zyplib.utils.ensure import ensure_3dims


def z_score(X: np.ndarray, axis: int = -1) -> np.ndarray:
    mean = np.mean(X, axis=axis, keepdims=True)
    std = np.std(X, axis=axis, keepdims=True)
    return (X - mean) / std


def minmax(
    X: np.ndarray, range_to: tuple[float, float] = (0, 1), axis: int = -1
) -> np.ndarray:
    min = np.min(X, axis=axis, keepdims=True)
    max = np.max(X, axis=axis, keepdims=True)
    return (X - min) / (max - min) * (range_to[1] - range_to[0]) + range_to[0]


def robust_scale(
    X: np.ndarray, percentiles: tuple[float, float] = (25, 75), axis: int = -1
) -> np.ndarray:
    median = np.median(X, axis=axis, keepdims=True)
    lower = np.percentile(X, percentiles[0], axis=axis, keepdims=True)
    upper = np.percentile(X, percentiles[1], axis=axis, keepdims=True)
    return (X - median) / (upper - lower)


def scale_eeg(
    signals: np.ndarray,
    method: Literal['minmax', 'robust', 'zscore'] = 'zscore',
    across: Literal['channel', 'epoch', 'total'] = 'channel',
    *,
    minmax_range: tuple[float, float] = (0, 1),
    robust_percentiles: tuple[float, float] = (25, 75),
) -> np.ndarray:
    """对 EEG 信号进行缩放（归一化）

    输入的 EEG 信号数据可能为 [C, T] 或 [N, C, T] 格式；

    指定缩放范围有三种，以下假设使用 [N, C, T] 格式和 minmax(0, 1) 缩放为例

    - `across='channel'`: 每个通道的信号直接的缩放互相独立; 即计算每个通道的 min 和 max，然后各自缩放到 [0, 1] 范围内
    - `across='epoch'`: 在完整的 [C, T] 的 Epoch 上进行缩放; 即计算 N 个 Epoch[C, T] 的 min 和 max，然后缩放到 [0, 1] 范围内
    - `across='total'`: 在完整的 [N, C, T] 的 Epoch 上进行缩放; 即计算整个 [N, C, T] 的 min 和 max，然后缩放到 [0, 1] 范围内

    Parameters
    ----------
    - `signals` : `np.ndarray`
        - 输入的 EEG 信号数据，形状为 (C, T) 或 (N, C, T)
    - `method` : `Literal['minmax', 'robust', 'zscore']`, 缩放方法
        - 'minmax' 将数据缩放到 [min, max] 范围内
        - 'zscore' 使用 z-score 标准化
        - 'robust' 使用中位数和四分位数进行缩放
    - `across` : `Literal['channel', 'epoch']`, 指定**缩放范围**
        - 'channel': 每个通道的信号直接的缩放互相独立
        - 'epoch': 在完整的 [C, T] 的 Epoch 上进行缩放
        - 'total': 在完整的 [N, C, T] 的 Epoch 上进行缩放
    - `minmax_range` : `tuple[float, float]`, optional, 仅在 `method='minmax'` 时有效
        - 缩放后的最小值和最大值，默认为 (0, 1)
    - `robust_percentiles` : `tuple[float, float]`, optional, 仅在 `method='robust'` 时有效
        - 缩放后的中位数和四分位数，默认为 (25, 75)

    """
    sig_shape = signals.shape
    signals = ensure_3dims(signals, newaxis='batch')

    N, C, T = signals.shape
    if method == 'zscore':
        scaler = z_score
    elif method == 'minmax':

        def fn_minmax(x, axis):
            return minmax(x, range_to=minmax_range, axis=axis)

        scaler = fn_minmax
    elif method == 'robust':

        def fn_robust(x, axis):
            return robust_scale(x, percentiles=robust_percentiles, axis=axis)

        scaler = fn_robust
    else:
        raise ValueError(f'Invalid method: {method}')

    if across == 'channel':
        # 对每个通道单独进行缩放
        scaled = scaler(signals, axis=-1)
    elif across == 'epoch':
        # 对每个 epoch 单独进行缩放
        # scaled = np.array([scaler(signals[i], axis=(0, 1)) for i in range(N)])
        scaled = scaler(signals, axis=(1, 2))
    elif across == 'total':
        # 对整个数据集进行缩放
        scaled = scaler(signals, axis=None)
    else:
        raise ValueError(f'Invalid across: {across}')

    return scaled.reshape(sig_shape)



import numpy as np
from scipy.signal import resample_poly


def resample_eeg(X: np.ndarray, fs: int, target_fs: int, axis: int = -1):
    """对 EEG 数据进行重采样; 将一个采样率为 `fs` 的 EEG 数据重采样为 `target_fs`

    考虑到 EEG 数据的非周期性, 使用 scipy.signal.resample_poly 进行重采样
    """
    return resample_poly(X, target_fs, fs, axis=axis)



"""
Matplotlib 相关的绘图工具
"""

from typing import Callable, Protocol

import matplotlib.pyplot as plt
import numpy as np

from ._style import register_mpl_styles


class MplFigureCallback(Protocol):
    """matplotlib 图表的最终处理函数类型; 比如可以在绘图后调用 `plt.show()` 等"""

    def __call__(self, fig: plt.Figure | None = None) -> None: ...


def mpl_figure_show(fig: plt.Figure | None = None):
    """显示 matplotlib 图表"""
    plt.show()


def make_grids(
    row: int = 1,
    col: int = 1,
    figsize: tuple = None,
    tight_layout: bool = True,
    fn_finalize: MplFigureCallback = mpl_figure_show,
    **subplots_kwargs,
):
    """创建 matplotlib 的子图网格，并逐个返回子图

    >>> for i, j, ax in make_grids(2, 3):
    ...     ax.plot(np.random.randn(100))

    >>> save_fig = lambda fig: fig.savefig('fig.png')
    >>> for i, j, ax in make_grids(2, 3, fn_finalize=save_fig):
    ...     ax.plot(np.random.randn(100))

    Parameters
    ----------
    - `row` : `int`, optional
        - 行数, 默认为 1
    - `col` : `int`, optional
        - 列数, 默认为 1
    - `figsize` : `tuple`, optional
        - 图形大小 (宽, 高), 单位为英寸, 默认为 None (使用 matplotlib 默认大小)
    - `tight_layout` : `bool`, optional
        - 是否使用 tight_layout, 默认为 True
    - `fn_finalize` : `MplFigureCallback`, optional
        - 图表的最终处理函数, 默认为 `mpl_figure_show`, 即简单地调用 `plt.show()`
    - `subplots_kwargs` : `dict`, optional
        - 其他参数, 传递给 `plt.subplots`

    Yields
    ----------
    - `int, int, plt.Axes`: `row`, `col`, `axes`
    """
    fig, axes = plt.subplots(row, col, figsize=figsize, **subplots_kwargs)

    if tight_layout:
        fig.tight_layout()

    # 确保 axes 始终是二维数组
    if row == 1 and col == 1:
        axes = np.array([[axes]])
    elif row == 1 or col == 1:
        axes = axes.reshape(row, col)

    for i in range(row):
        for j in range(col):
            yield i, j, axes[i, j]

    if fn_finalize:
        fn_finalize(fig)



"""
本模块主要提供了一些数据集导入的 pipeline 相关的函数，是对脑电信号数据导入、预处理流程的高级封装

所有函数均满足:

- 输入: 一些参数，以及一些回调函数
- 输出: 返回数据集 (X, y)

"""

from functools import reduce
from typing import Callable, Protocol, Sequence, runtime_checkable

import numpy as np
from sklearn.model_selection import KFold

from zyplib.dataset.labels import to_onehot
from zyplib.dataset.split import random_split

__all__ = ['basic_pipeline']


@runtime_checkable
class FuncPipeline(Protocol):
    """函数类型，输入文件路径列表，输出数据集 (X, y)"""

    def __call__(self, files: list[str]) -> tuple[np.ndarray, np.ndarray]: ...


@runtime_checkable
class ReadFunction(Protocol):
    def __call__(self, file_path: str) -> np.ndarray: ...


@runtime_checkable
class SegmentFunction(Protocol):
    def __call__(self, signal: np.ndarray) -> np.ndarray: ...


@runtime_checkable
class LabelFunction(Protocol):
    def __call__(self, file_path: str) -> int: ...


@runtime_checkable
class SignalProcessFunction(Protocol):
    def __call__(self, signal: np.ndarray) -> np.ndarray: ...


@runtime_checkable
class EncodeLabelFunction(Protocol):
    def __call__(self, label: list | np.ndarray) -> np.ndarray: ...


def basic_pipeline(
    files: list[str],
    fn_read: ReadFunction,
    fn_segment: SegmentFunction,
    label: LabelFunction,
    fn_before_segment: SignalProcessFunction = None,
    fn_after_segment: SignalProcessFunction = None,
    fn_encode_label: EncodeLabelFunction = lambda y: to_onehot(y, squeeze_2_class=True),
) -> tuple[np.ndarray, np.ndarray]:
    """从文件中读取数据，并自动切片、制作用于训练的数据集

    本 Pipeline 适用于:
    - 从文件中读取完整的脑电
    - 需要对完整脑电进行切分
    - 无特殊标签，只需根据文件进行标记即可
    - 输出: 数据集 (X, y)
        - X: 数据，shape = (N, C, T)
        - y: 标签，shape = (N,)

    Pipeline 流程:
    ----------

    1. 从文件中读取完整的数据 X
    2. 对数据 X 进行预处理
    3. 对数据 X 进行切片
    4. 对数据 X 进行后处理
    5. 制作标签 y
    6. 返回数据集 (X, y)

    Parameters
    ----------
    - `files` : `list[str]`
        - 文件路径列表
    - `fn_read` : `(file_path: str) -> np.ndarray`
        - 文件读取函数
    - `fn_segment` : `(signal: np.ndarray) -> np.ndarray`
        - 数据切片函数
        - 输入一个 `[C, T]` 的信号
        - 输出一个 `[N, C, T]` 的信号
    - `label` : `int | (file_path: str) -> int`
        - 标签，如果为 int，则所有数据标签相同；如果为函数，则根据文件名生成标签
    - `fn_before_segment` : `(signal: [C, T]) -> [C, T]`, optional
        - 数据切片前的处理函数
    - `fn_after_segment` : `(signal: [N, C, T]) -> [N, C, T]`, optional
        - 数据切片后的处理函数
    - `fn_encode_label` : `(label: list | np.ndarray) -> np.ndarray`, optional
        - 标签编码函数, 默认使用 `to_onehot`

    Returns
    ----------
    - `X` : `np.ndarray`, shape = (N, C, T)
        - 数据
    - `y` : `np.ndarray`, shape = (N, n_classes)
        - 标签
    """
    if not files:
        raise ValueError('The files list is empty. Please provide at least one file.')

    fn_before_segment = fn_before_segment or (lambda x: x)
    fn_after_segment = fn_after_segment or (lambda x: x)
    label = label if callable(label) else (lambda _: label)

    # 处理单个文件
    def load_single_file(fpath: str):
        signal = fn_read(fpath)  # 读取数据
        signal = fn_before_segment(signal)  # 数据预处理
        segments = fn_segment(signal)  # 数据切片
        segments = fn_after_segment(segments)
        N = len(segments)
        label_ = label(fpath)  # 制作标签
        return segments, [label_] * N

    # 批量处理所有文件
    results = map(load_single_file, files)

    # 合并数据
    def merge_data(Xy: tuple[list[np.ndarray], list[int]], pair: tuple[np.ndarray, int]):
        Xy[0].append(pair[0])
        Xy[1].extend(pair[1])
        return Xy

    Xy = reduce(merge_data, results, ([], []))

    X = np.concatenate(Xy[0], axis=0)
    y = np.array(Xy[1], dtype=int)

    fn_encode_label = fn_encode_label or (lambda y: to_onehot(y, squeeze_2_class=True))

    y = fn_encode_label(y)

    return X, y


def train_test_pipeline(
    train_size: float | int,
    all_files: list[str] | Sequence[list[str]],
    random_seed: int | None = None,
    return_split_files: bool = False,
    *,
    fn_pipeline: FuncPipeline | None = None,
    **basic_pipeline_kwargs,
):
    """在导入文件的时候，自动将不同文件划分为训练集和测试集

    - 需要一个 `fn_pipeline: (list[str]) -> (np.ndarray, np.ndarray)` 的回调函数
        - 输入: 文件路径列表
        - 输出: 数据集 (X, y)
    - 当 `fn_pipeline` 为 `None` 时，使用 `basic_pipeline`, 并且使用后面参数作为 `basic_pipeline` 的参数

    #### 使用自定义的 fn_pipeline

    ```py
    def load_xy(files: list[str]) -> tuple[np.ndarray, np.ndarray]:
        ...

    X_train, X_test, y_train, y_test = train_test_pipeline(
        0.8, all_files, fn_pipeline=load_xy
    )
    ```

    #### 默认使用 basic_pipeline
    ```py
    train_test_pipeline(
        0.8, all_files, fn_read=fn_read, fn_segment=fn_segment, label=label, ...
    )

    # 等价于
    train_test_pipeline(
        0.8, all_files, fn_pipeline=lambda files: basic_pipeline(files, ...)
    )
    ```

    Parameters
    ----------
    - `train_size` : `float | int`
        - 训练集大小
    - `all_files` : `list[str] | Sequence[list[str]]`
        - 所有文件路径列表
        - 如果为 `list[str]`，会直接将 `all_files` 划分为训练集和测试集
        - 如果为 `Sequence[list[str]]`，则视为不同分组的文件列表; 会各自单独划分为训练集和测试集最后再合并在一起
    - `random_seed` : `int | None`, optional
        - 随机种子; 当 `random_seed` 为 `None` 时，不设置随机种子
    - `return_split_files` : `bool`, optional
        - 是否返回训练集和测试集的文件路径列表; 当 `return_split_files` 为 `True` 时，返回训练集和测试集的文件路径列表
    - `fn_pipeline` : `FuncPipeline`, optional
        - 数据处理 pipeline; 当 `fn_pipeline` 为 `None` 时，使用 `basic_pipeline`, 并使用后面的参数作为 pipeline 的参数
    - `basic_pipeline_kwargs` `basic_pipeline` 相同, 包括:
        - `['fn_read', 'fn_segment', 'label', 'fn_before_segment', 'fn_after_segment']`

    Returns
    ----------
    - `X_train` : `np.ndarray`, shape = (N_train, C, T)
        - 训练数据
    - `X_test` : `np.ndarray`, shape = (N_test, C, T)
        - 测试数据
    - `y_train` : `np.ndarray`, shape = (N_train, n_classes)
        - 训练标签
    - `y_test` : `np.ndarray`, shape = (N_test, n_classes)
        - 测试标签
    - `train_files` : `list[str]`, optional
        - 训练集文件路径列表; 当 `return_split_files` 为 `True` 时返回
    - `test_files` : `list[str]`, optional
        - 测试集文件路径列表; 当 `return_split_files` 为 `True` 时返回
    """
    if isinstance(all_files[0], str):
        all_files = [all_files]  # Convert to a single group if it's a flat list

    train_files, test_files = [], []
    for group in all_files:
        N = len(group)
        if isinstance(train_size, int):
            group_train_size = min(train_size, N - 1)  # Ensure at least one test sample
        else:
            group_train_size = int(N * train_size)

        group_test_size = N - group_train_size

        group_train, group_test = random_split(
            group, (group_train_size, group_test_size), seed=random_seed
        )
        train_files.extend(group_train)
        test_files.extend(group_test)

    if fn_pipeline is None:
        fn_pipeline = lambda files: basic_pipeline(files, **basic_pipeline_kwargs)

    X_train, y_train = fn_pipeline(train_files)
    X_test, y_test = fn_pipeline(test_files)

    if return_split_files:
        return X_train, X_test, y_train, y_test, train_files, test_files
    else:
        return X_train, X_test, y_train, y_test


def kfold_pipeline(
    k: int,
    all_files: list[str],
    random_seed: int | None = None,
    yield_split_files: bool = False,
    *,
    fn_pipeline: FuncPipeline | None = None,
    **basic_pipeline_kwargs,
):
    """在导入文件的时候，自动将不同文件划分为 k 折, 通过 `yield` 返回每次的 XY 数据

    - 需要一个 `fn_pipeline: (list[str]) -> (np.ndarray, np.ndarray)` 的回调函数
        - 输入: 文件路径列表
        - 输出: 数据集 (X, y)
    - 当 `fn_pipeline` 为 `None` 时，使用 `basic_pipeline`, 并且使用后面参数作为 `basic_pipeline` 的参数

    Parameters
    ----------
    - `k` : `int`
        - 折数
    - `all_files` : `list[str]`
        - 所有文件路径列表
    - `random_seed` : `int | None`, optional
        - 随机种子; 当 `random_seed` 为 `None` 时，不设置随机种子
    - `yield_split_files` : `bool`, optional
        - 是否返回训练集和验证集的文件路径列表; 当 `yield_split_files` 为 `True` 时，返回训练集和验证集的文件路径列表
    - `fn_pipeline` : `FuncPipeline`, optional
        - 数据处理 pipeline; 当 `fn_pipeline` 为 `None` 时，使用 `basic_pipeline`, 并使用后面的参数作为 pipeline 的参数
    - `basic_pipeline_kwargs` `basic_pipeline` 相同, 包括:
        - `['fn_read', 'fn_segment', 'label', 'fn_before_segment', 'fn_after_segment']`

    Yields
    ----------
    - `X_train` : `np.ndarray`, shape = (N_train, C, T)
        - 训练数据
    - `X_val` : `np.ndarray`, shape = (N_val, C, T)
        - 验证数据
    - `y_train` : `np.ndarray`, shape = (N_train, n_classes)
        - 训练标签
    - `y_val` : `np.ndarray`, shape = (N_val, n_classes)
        - 验证标签
    - `train_files` : `list[str]`, optional
        - 训练集文件路径列表; 当 `yield_split_files` 为 `True` 时返回
    - `val_files` : `list[str]`, optional
        - 验证集文件路径列表; 当 `yield_split_files` 为 `True` 时返回
    """
    kf = KFold(n_splits=k, shuffle=True, random_state=random_seed)

    if fn_pipeline is None:
        fn_pipeline = lambda files: basic_pipeline(files, **basic_pipeline_kwargs)

    for train_idx, val_idx in kf.split(all_files):
        train_files = [all_files[i] for i in train_idx]
        val_files = [all_files[i] for i in val_idx]

        X_train, y_train = fn_pipeline(train_files)
        X_val, y_val = fn_pipeline(val_files)

        if yield_split_files:
            yield X_train, X_val, y_train, y_val, train_files, val_files
        else:
            yield X_train, X_val, y_train, y_val



from pathlib import Path
from typing import Literal, Protocol

import numpy as np

from zyplib.annotation.event import Event


class EEGFileInterface(Protocol):
    """EEG 文件接口"""

    filepath: Path  # 信号文件路径
    meta: dict  # 信号的元数据
    events: list[Event]  # 信号的标注
    name: str  # 信号的名称
    channels: list[str]  # 信号通道标签, 长度等于 `n_channels`
    n_channels: int  # 信号的通道数量
    srate: float  # 信号的采样率
    t_duration: float  # 信号的时间长度; 单位: 秒
    n_samples: int  # 信号的时间样本数量; n_samples = t_duration * sample_rate
    data: np.ndarray  # 信号数据; 形状为 (n_channels, n_samples)
    units: dict[str, Literal['V', 'uV', 'mV', '?']]

    def get_epoch(self, t_start: float, t_duration: float) -> np.ndarray:
        """获取指定时间范围内的信号数据

        >>> signals.get_epoch(t_start=0, t_duration=1)
        >>> # 获取从0秒到1秒的信号数据

        Parameters
        ----------
        - `t_start` : `float`
            - 开始时间; 单位: 秒
        - `t_duration` : `float`
            - 时间长度; 单位: 秒

        Returns
        ----------
        - `np.ndarray`
            - 信号数据; 形状为 (n_channels, n_samples)
        """

    def get_events(
        self, t_start: float, t_duration: float | None = None
    ) -> list[Event] | None:
        """获取指定时间范围内的标注"""
        ...

    def add_event(self, event: Event): ...

    def delete_event(self, t_onset: float, t_duration: float | None = None): ...

    def pick_event(self, t_at: float, episilon: float = 0.1) -> list[Event]: ...


class EventMixin:
    def __init__(self, events: list[Event] | None = None) -> None:
        events = events or []
        self._events: list[Event] = events

    @property
    def events(self) -> list[Event]:
        return self._events

    def _sort_events(self):
        self._events.sort(key=lambda x: x.onset)

    def get_events(self, t_start: float, t_duration: float | None = None) -> list[Event]:
        self._sort_events()
        events = []
        for event in self._events:
            if t_start <= event.onset <= t_start + t_duration:
                events.append(event)
        return events

    def pick_event(self, t_at: float, episilon: float = 0.1) -> list[Event]:
        """获取在 t_at 附近的标注


        Parameters
        ----------
        - `t_at` : `float`
            - 时间点
        - `episilon` : `float`
            - 时间精度

        Returns
        ----------
        - `list[Event]`
            - 在 t_at 附近的标注
        """
        self._sort_events()
        events = []
        for event in self._events:
            if event.duration is None or event.duration <= episilon:
                if t_at - episilon <= event.onset <= t_at + episilon:
                    events.append(event)
            else:
                if (
                    event.onset - episilon
                    <= t_at
                    <= event.onset + event.duration + episilon
                ):
                    events.append(event)
        return events

    def add_event(self, event: Event):
        if event in self._events:
            return False
        self._events.append(event)

    def delete_event(self, event: Event):
        self._events.remove(event)



import time
from datetime import datetime, timedelta, timezone
from timeit import default_timer as timer
from typing import Optional

from .print import colored


class TIME_FMT:
    YEAR = '%Y'
    YEAR_SHORT = '%y'
    MONTH = '%m'
    DAY = '%d'
    HOUR = '%H'
    MINUTE = '%M'
    SECOND = '%S'

    DATE = f'{YEAR}-{MONTH}-{DAY}'
    TIME = f'{HOUR}:{MINUTE}:{SECOND}'
    DATE_TIME = f'{DATE} {TIME}'
    DATE_TIME_CONDENSED = f'{YEAR}{MONTH}{DAY}{HOUR}{MINUTE}{SECOND}'
    DATE_TIME_UNDERSCORE = f'{YEAR}_{MONTH}_{DAY}-{HOUR}_{MINUTE}_{SECOND}'


def format_time(ts: float, fmt: str = TIME_FMT.DATE_TIME, time_zone: int = 8):
    """格式化时间为字符串

    Parameters
    ----------
    - `ts` : `float`
        - 时间戳, time.time() 的返回值
    - `fmt` : `str`, optional
        - 格式化字符串，默认为 `TIME_FMT.DATE_TIME`
    - `time_zone` : `int`, optional
        - 时区，默认为东八区

    Returns
    ----------
    - `str`
        - 格式化后的时间字符串
    """
    td = timedelta(hours=time_zone)
    tz = timezone(td)
    dt = datetime.fromtimestamp(ts, tz)
    dt = dt.strftime(fmt)
    return dt


def format_time_delta(seconds: float) -> str:
    delta = timedelta(seconds=seconds)
    return str(delta)[:-3]


def now(fmt: str = TIME_FMT.DATE_TIME, time_zone: int = 8) -> str:
    """获取当前时间

    >>> print(now())
    2024-10-12 16:00:00

    Parameters
    ----------
    - `fmt` : `str`, optional
        - 格式化字符串，默认为 `TIME_FMT.DATE_TIME`
    - `time_zone` : `int`, optional
        - 时区，默认为东八区

    Returns
    ----------
    - `str`
        - 当前时间字符串
    """
    ts = time.time()
    return format_time(ts, fmt, time_zone)


class TicToc:
    def __init__(self):
        self._tic = 0
        self._toc = 0
        self._elapsed = 0

    def tic(self):
        self._tic = timer()

    def toc(self):
        self._toc = timer()
        self._elapsed = self._toc - self._tic

    @property
    def elapsed(self):
        delta = timedelta(seconds=self._elapsed)
        return str(delta)[:-3]

    def print_elapsed(self, name: Optional[str] = None):
        if name is None:
            name = 'It'
        else:
            name = colored(name, 'yellow')
        print(f'{name} took {colored(self.elapsed, "cyan")}')


def print_elapsed(func):
    """打印函数执行时间

    ```python
    @print_elapsed
    def my_func():
        import time
        time.sleep(1)
    ```

    >>> my_func()
    my_func took 0:00:01.007 seconds

    Parameters
    ----------
    - `func` : `Callable`
        - 被装饰的函数

    Returns
    ----------
    - `Callable`
        - 装饰后的函数
    """

    def wrapper(*args, **kwargs):
        tictoc = TicToc()
        tictoc.tic()
        result = func(*args, **kwargs)
        tictoc.toc()
        tictoc.print_elapsed(func.__name__)
        return result

    return wrapper



from typing import Any, Literal


def type_summary(obj: Any) -> str:
    """返回对象的类型摘要

    此函数接受任何类型的对象，并返回一个字符串，描述该对象的类型和一些基本信息。

    Examples
    --------

    >>> type_summary("hello")
    'hello'
    >>> type_summary(42)
    '42'
    >>> type_summary(3.14159)
    '3.142'
    >>> type_summary([1, 2, 3])
    'int[3]'
    >>> type_summary({'a': 1, 'b': 2})
    "dict[str: int]"
    >>> import numpy as np
    >>> type_summary(np.array([1, 2, 3]))
    'ndarray@int64(3,)'
    """
    if isinstance(obj, (str, int, bool)):
        return obj
    elif isinstance(obj, float):
        return f'{obj:.3f}'
    elif isinstance(obj, list):
        if len(obj) == 0:
            return '[]'
        else:
            # int[3]
            item_type = type(obj[0])
            return f'{item_type.__name__}[{len(obj)}]'
    elif isinstance(obj, tuple):
        return tuple([type_summary(k) for k in obj])
    elif isinstance(obj, dict):
        if len(obj) == 0:
            return f'{type(obj).__name__}[]'
        else:
            key_type = type(next(iter(obj.keys())))
            value_type = type(next(iter(obj.values())))
            return f'{type(obj).__name__}[{key_type.__name__}: {value_type.__name__}]'
    elif hasattr(obj, 'shape') and hasattr(obj, 'dtype'):
        return f'{type(obj).__name__}@{obj.dtype}{obj.shape}'
    else:
        return type(obj).__name__


COLOR_MAP = {
    'red': '\033[91m',
    'green': '\033[92m',
    'yellow': '\033[93m',
    'blue': '\033[94m',
    'magenta': '\033[95m',
    'cyan': '\033[96m',
    'white': '\033[97m',
    'gray': '\033[90m',
    'reset': '\033[0m',
}


COLOR_TYPE = Literal['red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'gray']


def colored(text: str, color: COLOR_TYPE):
    """为文本添加颜色


    Parameters
    ----------
    - `text` : `str`
        - 原始文本
    - `color` : `COLOR_TYPE`
        - 颜色类型
        - `"red"` | `"green"` | `"yellow"` | `"blue"` | `"magenta"` | `"cyan"` | `"white"` | `"gray"`
    """
    return f"{COLOR_MAP[color]}{text}{COLOR_MAP['reset']}"


def print_colored(color: COLOR_TYPE, *args, **kwargs):
    """打印带颜色的文本

    Parameters
    ----------
    - `color` : `COLOR_TYPE`
        - 颜色类型
        - `"red"` | `"green"` | `"yellow"` | `"blue"` | `"magenta"` | `"cyan"` | `"white"` | `"gray"`
    """

    total_text = ' '.join([str(arg) for arg in args])
    print(colored(total_text, color), **kwargs)


def print_debug(*args, **kwargs):
    """打印调试信息（白色）"""
    print_colored('gray', *args, **kwargs)


def print_info(*args, **kwargs):
    """打印信息（青色）"""
    print_colored('cyan', *args, **kwargs)


def print_warn(*args, **kwargs):
    """打印警告信息（黄色）"""
    print_colored('yellow', *args, **kwargs)


def print_warning(*args, **kwargs):
    """打印警告信息（黄色）"""
    print_colored('yellow', *args, **kwargs)


def print_error(*args, **kwargs):
    """打印错误信息（红色）"""
    print_colored('red', *args, **kwargs)



from typing import Any, Dict, List, Literal, Optional, Tuple, Union
from dataclasses import dataclass, field
import re

import h5py
import mne
import numpy as np

from .misc import match_channels_regex
from zyplib.annotation.event import Event, events_to_dict


@dataclass
class StandardData:
    Data: List[np.ndarray] = field(default_factory=list)
    Channels: List[str] = field(default_factory=list)
    TimeLength: float = -1
    Srate: List[float] = field(default_factory=list)
    Events: List[Event] = field(default_factory=list)
    Description: str = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._is_std_type = False

    def _as_std_type_(self):
        """按照标准格式来存储

        1. Channels: list，存储通道信息
        2. Data: list or ndarry，存储数据
            - 按照 float32 格式存储
            - 采样率最大不超过 256Hz
        3. Events：事件标签信息
        4. Srate: list, 各个通道的采样率
        5. Class：文件类型，数据or标签
        """
        if self._is_std_type:
            return

        if isinstance(self.Channels, np.ndarray):
            self.Channels = self.Channels.tolist()

        # 保证 Srate 是一个 list of int
        if isinstance(self.Srate, np.ndarray):
            self.Srate = self.Srate.squeeze().astype(int).tolist()
        if isinstance(self.Srate, (int, float)):
            self.Srate = [int(self.Srate)] * len(self.Channels)

        # 如果 Data 为 ndarray, 且为 float64, 则变为 float32
        if isinstance(self.Data, np.ndarray):
            self.Data = self.Data.astype(np.float32)
        elif isinstance(self.Data, list) and isinstance(self.Data[0], np.ndarray):
            for i in range(len(self.Channels)):
                data = self.Data[i].astype(np.float32)
                self.Data[i] = data

        if self.Data is not None:
            self.TimeLength = len(self.Data[0]) / self.Srate[0]

        self._is_std_type = True

    def dump_h5(self, path):
        self._as_std_type_()
        with h5py.File(path, 'w') as f:
            # 保存元数据字段到Group
            group = f.create_group('meta')
            group.attrs['Channels'] = self.Channels if self.Channels is not None else []
            group.attrs['Srate'] = self.Srate if self.Srate is not None else []
            group.attrs['TimeLength'] = self.TimeLength
            group.attrs['Description'] = self.Description

            # 保存data字段到Dataset
            data_group = f.create_group('Data')
            if self.Data is not None:
                for i, ndarr in enumerate(self.Data):
                    data_group.create_dataset(str(i), data=ndarr)

            events_group = f.create_group('Events')
            if not self.Events:
                dic = events_to_dict(self.Events)
                events_group.create_dataset('Onset', data=dic['Onset'])
                events_group.create_dataset('Duration', data=dic['Duration'])
                events_group.create_dataset('Name', data=dic['Name'])

    @staticmethod
    def load_h5(
        path, meta_only=False, match_channels: List[Union[str, re.Pattern]] = None
    ) -> 'StandardData':
        """读取 std-h5 文件

            Parameters
            ----------
                - `path` : str\\
                    h5 文件的路径
                - `meta_only` : bool, optional\\
                    是否只读取元数据
                - `match_channels` : List[Union[str, re.Pattern]], optional\\
                    正则表达式列表, 见 match_channels_regex

            Returns
            ----------
                - `StandardData`\\
                    _description_
        """
        with h5py.File(path, 'r') as f:
            # 读取元数据字段
            meta = f['meta']
            std_data = StandardData()
            std_data.Channels: List = meta.attrs['Channels'].tolist()
            std_data.Srate = meta.attrs['Srate'].tolist()
            std_data.TimeLength = meta.attrs['TimeLength']
            std_data.Description = meta.attrs['Description']

            events_group = f['Events']
            if len(events_group.keys()) > 0:
                dic = {}
                dic['onset'] = events_group['Onset'][()]
                dic['duration'] = events_group['Duration'][()]
                dic['name'] = events_group['Name'][()]
                std_data.Events.append(Event(**dic))

            if match_channels is not None:
                matched_ch_names, matched_ch_index = match_channels_regex(
                    std_data.Channels, match_channels
                )
                std_data.Channels = matched_ch_names
                std_data.Srate = [std_data.Srate[i] for i in matched_ch_index]
                include_idx = matched_ch_index
            else:
                include_idx = [i for i in range(len(std_data.Channels))]

            # 读取data字段
            if not meta_only:
                data_group = f['Data']
                std_data.Data = []
                for i in include_idx:
                    std_data.Data.append(data_group[str(i)][()])

        return std_data

    @property
    def data_array(self) -> np.ndarray:
        """将 Data 转换为 numpy.ndarray"""
        return np.array(self.Data)

    @property
    def is_srate_indentical(self) -> bool:
        """所有通道的采样率是否一致"""
        return len(self.Srate) == 0 or len(set(self.Srate)) == 1

    def to_mne_raw(self):
        """转换为 mne.io.Raw 对象"""
        channels, data, srate = self.Channels, self.Data, self.Srate

        if not self.is_srate_indentical:
            raise ValueError('All channels must have same srate')

        sr = srate[0] if len(srate) > 0 else 1
        info = mne.create_info(channels, sr, ch_types='eeg', verbose=False)
        raw = mne.io.RawArray(data, info)
        if len(self.Events) > 0:
            events_dict = events_to_dict(self.Events)
            raw.set_annotations(
                mne.Annotations(
                    onset=events_dict['Onset'],
                    duration=events_dict['Duration'],
                    description=events_dict['Name'],
                )
            )
        return raw



"""
文件系统相关工具
"""

import json
import os
import shutil
from pathlib import Path
from typing import Any, Literal, Union

import yaml

from .print import print_debug

FilePath = Union[Path, str]


def expand_user(path: FilePath) -> FilePath:
    return Path(path).expanduser()


def expand_relative(path: FilePath) -> FilePath:
    """完全基于调用文件位置而非工作目录的相对路径

    >>> expand_relative('../data/')  # 在 /home/ 下运行的 /home/user/a/test.py 文件
    >>> /home/user/data/
    """
    file_dir = Path(__file__).parent
    return file_dir.joinpath(path).resolve()


def ensure_dir(
    dir: str,
    if_not_empty: Literal['keep', 'clean', 'error'] = 'keep',
    verbose: bool = False,
):
    """检查用于保存日志、Checkpoint 等的目录如果 dir 不存在，则创建

    Parameters
    ----------
    - `dir` : `str`
        - 目录
    - `if_not_empty` : `Literal['keep', 'clean', 'error']`, optional
        - 如果目录不为空，则执行的操作
        - `"keep"` | `"clean"` | `"error"`
    - `verbose` : `bool`, optional
        - 是否打印日志，默认为 False
    """
    dir = str(dir).strip()
    if not dir or dir == '':
        return
    if not os.path.exists(dir):
        os.makedirs(dir)
    elif any(os.scandir(dir)):
        if if_not_empty is None or if_not_empty == 'keep':
            if verbose:
                print_debug(f'ensure_dir: exists {dir}.')
        elif if_not_empty == 'clean':
            if verbose:
                print_debug(f'ensure_dir: clean up {dir}.')
            shutil.rmtree(dir)
            os.makedirs(dir)
        elif if_not_empty == 'error':
            raise Exception(f'目录 {dir} 不为空!')
        else:
            raise ValueError(f'check_dir 参数错误: if_not_empty = {if_not_empty}')


def file_reader(file_path: FilePath, encoding: str = 'utf-8', binary: bool = False):
    mode = 'rb' if binary else 'r'
    return open(file_path, mode, encoding=None if binary else encoding)


def read_file(
    file_path: FilePath,
    encoding: str = 'utf-8',
    binary: bool = False,
    readlines: bool = False,
) -> Union[str, list[str]]:
    with file_reader(file_path, encoding, binary) as f:
        if readlines:
            return f.readlines()
        return f.read()


def file_writer(file_path: FilePath, encoding: str = 'utf-8', binary: bool = False):
    mode = 'wb' if binary else 'w'
    return open(file_path, mode, encoding=None if binary else encoding)


def write_file(
    file_path: FilePath,
    content: Union[str, bytes],
    encoding: str = 'utf-8',
    binary: bool = False,
):
    ensure_dir(os.path.dirname(file_path))
    with file_writer(file_path, encoding, binary) as f:
        f.write(content)


def read_json(file_path: FilePath, encoding: str = 'utf-8'):
    with open(file_path, 'r', encoding=encoding) as f:
        return json.load(f)


def write_json(file_path: FilePath, data: dict | list, encoding: str = 'utf-8'):
    ensure_dir(os.path.dirname(file_path))
    with open(file_path, 'w', encoding=encoding) as f:
        json.dump(data, f, ensure_ascii=False, indent=2)


# # 自定义标签处理函数，将其转换为字符串
# def ignore_unknown_tag(loader, node):
#     return loader.construct_scalar(node)


# # 注册 pathlib.PosixPath 的处理方式为忽略
# yaml.SafeLoader.add_constructor(
#     'tag:yaml.org,2002:python/object/apply:pathlib.PosixPath', ignore_unknown_tag
# )


def read_yaml(file_path: FilePath, throw_error: bool = True):
    try:
        with open(file_path, 'r', encoding='utf-8') as f:
            return yaml.safe_load(f)
    except yaml.YAMLError as e:
        if throw_error:
            raise e
        return {}


# 定义字符串呈现器（处理多行字符串）
def str_presenter(dumper, data):
    if len(data.splitlines()) > 1:  # check for multiline string
        return dumper.represent_scalar('tag:yaml.org,2002:str', data, style='|')
    return dumper.represent_scalar('tag:yaml.org,2002:str', data)


# 注册 str 类型的全局呈现器（只需要注册一次）
yaml.add_representer(str, str_presenter)


def write_yaml(file_path: FilePath, data: dict):
    # 确保导出时只使用基本的数据类型
    def dump_for_yaml(obj: Any) -> Any:
        if isinstance(obj, (int, float, bool, str)):
            return obj
        elif isinstance(obj, dict):
            return {k: dump_for_yaml(v) for k, v in obj.items()}
        elif isinstance(obj, (list, tuple, set)):
            return [dump_for_yaml(i) for i in obj]
        elif obj is None:
            return None
        elif hasattr(obj, 'to_dict'):  # 支持自定义对象的序列化
            return dump_for_yaml(obj.to_dict())
        return str(obj)


    clean_data = dump_for_yaml(data)  # 清理自定义标签

    ensure_dir(os.path.dirname(file_path))  # 确保目录存在
    with open(file_path, 'w', encoding='utf-8') as f:
        yaml.dump(clean_data, f, default_flow_style=False, allow_unicode=True)



import matplotlib.pyplot as plt
import mne
import numpy as np

from zyplib.signal.psd import welch


def topmap(
    data: list[float] | np.ndarray,
    channels: list[str],
    srate: float,
    montage_type='standard_1020',
    show=False,
    save_path=None,
    args_tompomap: dict = None,
):
    """绘制头皮数据的拓扑图。

    Parameters
    ----------
    - `data` : list[float] | np.ndarray
        - 拓扑图上绘制的数据。这应该是一个一维的值数组，每个通道一个值
    - `channels` : list[str]
        - 与数据对应的通道名称列表；长度和 data 长度相同
    - `srate` : float
        - 数据的采样率，单位为 Hz
    - `montage_type` : str, optional
        - 要使用的脑电图 Montage 类型，传入；默认为 'standard_1020'
        - 见 `mne.channels.get_builtin_montages`
    - `show` : bool, optional
        - 是否显示绘图。默认为 False。
    - `save_path` : str, optional
        - 保存绘图的路径。如果为 None，则不保存绘图。默认为 None。
    - `args_tompomap` : dict, optional
        - 要传递给 `mne.viz.plot_topomap` 的其他关键字参数。默认为空。
    """
    info = mne.create_info(ch_names=channels, sfreq=srate, ch_types='eeg')
    montage = mne.channels.make_standard_montage(montage_type)
    info.set_montage(montage)

    mne.viz.plot_topomap(data, info, show=show, cmap='viridis', **args_tompomap)

    if save_path:
        plt.savefig(save_path, format='png', transparent=True)


def pxx_topomap(
    signals: np.ndarray,
    channels: list[str],
    srate: float,
    montage_type='standard_1020',
    show=False,
    save_path=None,
    nperseg=None,
    args_tompomap: dict = None,
):
    """绘制脑电数据功率谱的拓扑图

    提取 signals 的功率谱，并绘制拓扑图

    Parameters
    ----------
    - `signals` : `np.ndarray`
        - 二维脑电信号 `[C, T]`
    - `channels` : list[str]
        - 与数据对应的通道名称列表；长度和 data 长度相同
    - `srate` : float
        - 数据的采样率，单位为 Hz
    - `montage_type` : str, optional
        - 要使用的脑电图 Montage 类型，传入；默认为 'standard_1020'
        - 见 `mne.channels.get_builtin_montages`
    - `show` : bool, optional
        - 是否显示绘图。默认为 False。
    - `save_path` : str, optional
        - 保存绘图的路径。如果为 None，则不保存绘图。默认为 None。
    - `nperseg` : int, optional
        - Welch 窗长。如果不填写默认为 2 * srate，即 2s
    - `args_tompomap` : dict, optional
        - 要传递给 `mne.viz.plot_topomap` 的其他关键字参数。默认为空。
    """
    nperseg = nperseg or 4 * srate
    _, pxx = welch(signals, srate, nperseg=2 * srate)

    # 计算每个通道的平均功率谱
    psds_mean = np.mean(pxx, axis=1)

    topmap(
        psds_mean,
        channels,
        srate,
        montage_type=montage_type,
        show=show,
        save_path=save_path,
        args_tompomap=args_tompomap,
    )



import json
import os
from dataclasses import asdict, dataclass

BASE_DIR = os.path.expanduser('~/.zyplib')
CONFIG_PATH = os.path.join(BASE_DIR, 'config.json')


@dataclass
class Config:
    DISK_CACHE_DIR: str = './cache'
    DISK_CACHE_MAX_SIZE: int = 3 * 1024 # MB

    def asdict(self):
        return asdict(self)


def _write_config(config: Config):
    """写入全局的配置文件"""
    if not os.path.exists(BASE_DIR):
        os.makedirs(BASE_DIR)

    config_path = os.path.expanduser(CONFIG_PATH)
    with open(config_path, 'w', encoding='utf-8') as file:
        json.dump(config.__dict__, file, indent=4)


def _load_config() -> Config:
    # Load global config
    try:
        global_config = _load_global_config()
    except Exception:
        global_config = Config()

    # Load local config and merge with global config
    try:
        local_config = _load_local_config()
    except Exception:
        local_config = {}
    merged_config = {**asdict(global_config), **local_config}

    # Create final config object
    config = Config(**merged_config)
    return config


def _load_global_config() -> Config:
    config_path = os.path.expanduser(CONFIG_PATH)
    if os.path.exists(config_path):
        with open(config_path, 'r', encoding='utf-8') as file:
            config_dict = json.load(file)
        return Config(**config_dict)
    else:
        _write_config(Config())
    return Config()


def _load_local_config() -> dict:
    local_config_path = 'zyplib.config.json'
    if os.path.exists(local_config_path):
        with open(local_config_path, 'r', encoding='utf-8') as file:
            return json.load(file)
    return {}


# 实例化配置，在模块导入时加载
config = _load_config()



"""
cnn+rnn for 1-d signal data, pytorch version

Shenda Hong, Jan 2020
"""

import torch
import torch.nn as nn


class CRNN(nn.Module):
    def __init__(
        self, in_channels: int, out_channels: int, n_samples: int, n_classes: int
    ):
        """一个简单的一维卷积神经网络+LSTM的模型

        @source: https://github.com/hsd1503/resnet1d/blob/master/crnn1d.py

        - Input: (batch, channels, length)
        - Output: (batch)
        """
        super(CRNN, self).__init__()

        self.n_samples = n_samples
        self.n_classes = n_classes
        self.in_channels = in_channels
        self.out_channels = out_channels

        # (batch, channels, length)
        self.cnn = nn.Conv1d(
            in_channels=self.in_channels,
            out_channels=self.out_channels,
            kernel_size=16,
            stride=2,
        )
        # (batch, seq, feature)
        self.rnn = nn.LSTM(
            input_size=(self.out_channels),
            hidden_size=self.out_channels,
            num_layers=1,
            batch_first=True,
            bidirectional=False,
        )
        self.dense = nn.Linear(out_channels, n_classes)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.n_channel, self.n_length = x.shape[-2], x.shape[-1]
        assert (
            self.n_length % self.n_samples == 0
        ), 'Input n_length should be divisible by n_len_seg'
        self.n_seg = self.n_length // self.n_samples

        out = x.permute(0, 2, 1)
        out = out.view(-1, self.n_samples, self.n_channel)
        out = out.permute(0, 2, 1)
        out = self.cnn(out)
        out = out.mean(-1)
        out = out.view(-1, self.n_seg, self.out_channels)
        _, (out, _) = self.rnn(out)
        out = torch.squeeze(out, dim=0)
        out = self.dense(out)

        return out



import os
from typing import Any, Dict, Optional, Union

import yaml
from lightning.fabric.loggers import Logger
from typing_extensions import override


class LoggerPropertyMixin:
    @property
    @override
    def name(self) -> str:
        """Gets the name of the experiment.

        Returns:
            The name of the experiment.

        """
        return self._name

    @property
    @override
    def version(self) -> Union[int, str]:
        """Gets the version of the experiment.

        Returns:
            The version of the experiment if it is specified, else the next version.

        """
        if self._version is None:
            self._version = 0
        return self._version

    @property
    @override
    def root_dir(self) -> str:
        """Gets the save directory where the versioned CSV experiments are saved."""
        return self._root_dir

    @property
    @override
    def log_dir(self) -> str:
        """The log directory for this run.

        By default, it is named ``'version_${self.version}'`` but it can be overridden by passing a string value for the
        constructor's version parameter instead of ``None`` or an int.

        """
        # create a pseudo standard path
        version = (
            self.version if isinstance(self.version, str) else f'version_{self.version}'
        )
        return os.path.join(self._root_dir, self.name, version)


class YAMLLogger(LoggerPropertyMixin, Logger):
    def __init__(
        self,
        root_dir: str = './',
        name: str = 'lightning_logs',
        version: Optional[Union[int, str]] = None,
        save_every_k: Optional[int] = 100,
    ):
        LoggerPropertyMixin.__init__(self)
        Logger.__init__(self)
        self._root_dir = root_dir
        self._name = name
        self._version = version
        self._log_file = os.path.join(self.log_dir, 'metrics.yaml')
        self._save_every_k = save_every_k
        self._log_buffer = []

        # Create the log directory if it doesn't exist
        os.makedirs(self.log_dir, exist_ok=True)

        self._auto_step = {}

    @override
    def log_hyperparams(self, params: Dict[str, Any]) -> None:
        with open(self._log_file, 'a', encoding='utf-8') as f:
            yaml.dump({'hyperparameters': params}, f)

    @override
    def log_metrics(self, metrics: Dict[str, float], step: Optional[int] = None) -> None:
        # Implement auto step functionality
        if step is None:
            metric_hash = '&'.join(sorted(metrics.keys()))
            if metric_hash not in self._auto_step:
                self._auto_step[metric_hash] = 0
            step = self._auto_step[metric_hash]
            self._auto_step[metric_hash] += 1

        log_entry = {'step': step, 'metrics': metrics}
        self._log_buffer.append(log_entry)

        if self._save_every_k is not None and len(self._log_buffer) >= self._save_every_k:
            self._write_buffer_to_file()

    def _write_buffer_to_file(self) -> None:
        with open(self._log_file, 'a', encoding='utf-8') as f:
            yaml.dump(self._log_buffer, f)
        self._log_buffer.clear()

    @override
    def save(self) -> None:
        self._write_buffer_to_file()

    @override
    def finalize(self, status: str) -> None:
        self._write_buffer_to_file()
        log_entry = {'status': status}
        with open(self._log_file, 'a', encoding='utf-8') as f:
            yaml.dump([log_entry], f)



import os

import matplotlib.pyplot as plt


def register_mpl_styles(plt_use: str | list[str] | None = None):
    """
    注册 matplotlib 内置的样式

    源码
    ----------
    本代码来自: https://github.com/garrettj403/SciencePlots/blob/master/scienceplots/__init__.py
    """
    current_dir_path = os.path.dirname(os.path.abspath(__file__))
    styles_path = os.path.join(current_dir_path, 'styles')

    # Reads styles in /styles folder and all subfolders
    stylesheets = {}  # plt.style.library is a dictionary
    for folder, _, _ in os.walk(styles_path):
        new_stylesheets = plt.style.core.read_style_directory(folder)
        stylesheets.update(new_stylesheets)

    # Update dictionary of styles - plt.style.library
    plt.style.core.update_nested_dict(plt.style.library, stylesheets)
    # Update `plt.style.available`, copy-paste from:
    # https://github.com/matplotlib/matplotlib/blob/a170539a421623bb2967a45a24bb7926e2feb542/lib/matplotlib/style/core.py#L266  # noqa: E501
    plt.style.core.available[:] = sorted(plt.style.library.keys())

    if plt_use is not None:
        plt.style.use(plt_use)



from typing import Protocol

import numpy as np
import pandas as pd

from zyplib.utils.ensure import ensure_3dims, ensure_npy, ensure_scalar


# @runtime_checkable
class EEGFeatureFunction(Protocol):
    """输入脑电信号 [N, C, T], 输出:

    1. 标量
    2. 向量 [N,]
    3. 矩阵 [N, C] 或 [N, K]
    4. 三维张量 [N, C, K]
    """

    def __call__(self, signal: np.ndarray) -> np.ndarray | float: ...


def make_tabular(
    signals: np.ndarray,
    fn_features: list[EEGFeatureFunction] | dict[str, EEGFeatureFunction],
    return_feat_names: bool = False,
) -> np.ndarray | tuple[np.ndarray, list[str]]:
    """将脑电信号转换为表格数据

    此函数接收脑电信号数据和一系列特征提取函数，将每个特征函数应用于输入信号，
    并将结果组合成一个表格形式的特征矩阵。

    Parameters
    ----------
    - `signals` : `np.ndarray`
        - 脑电信号, 形状为 (n_samples, n_channels, n_times)
        - 如果输入是2维数组, 将被视为 `(1, n_channels, n_times)`
    - `fn_features` : `list[EEGFeatureFunction] | dict[str, EEGFeatureFunction]`， 特征函数, 可以是以下两种形式之一
        1. 函数列表：每个函数应接受形状为 [n_samples, n_channels, n_times] 的 numpy 数组，
           并返回标量、一维数组或与输入兼容的多维数组
        2. 字典：键为特征名称，值为特征函数
    - `return_feat_names` : `bool`, optional
        - 是否返回特征名称, 默认为 False

    Returns
    ----------
    - `np.ndarray | tuple[np.ndarray, list[str]]`
        - 如果 `return_feat_names` 为 False：
            返回表格数据, 形状为 (n_samples, n_features)
        - 如果 `return_feat_names` 为 True：
            返回一个元组, 包含表格数据和特征名称列表

    Notes
    ----------
    1. 特征函数的返回值处理：
       - 标量：被广播到所有样本
       - 一维数组 `[N,]`：必须与样本数量匹配，每个元素对应一个样本的特征
       - 二维数组 `[N, K]` 或 `[N, C]`：
         * 如果第二维等于通道数，假定为每个通道的特征
         * 否则，假定为每个样本的 K 个特征
       - 三维数组 `[N, C, K]`：被解释为每个样本、每个通道的 K 个特征

    2. 特征名称生成规则：
       - 标量或一维数组：使用函数名或字典键
       - 二维数组：
         * 通道特征：`'name_ch{j}'`，j 为通道索引
         * 多特征：`'name{i}'`，i 为特征索引
       - 三维数组：`'name{i}_ch{j}'`，i 为特征索引，j 为通道索引

    Examples
    ----------
    >>> import numpy as np
    >>> from zyplib.feature.tabular import make_tabular
    >>>
    >>> # 创建示例数据
    >>> signals = np.random.rand(10, 4, 1000)  # 10 个样本，4 个通道，1000 个时间点
    >>>
    >>> # 定义一些特征函数
    >>> def mean_feature(x):
    ...     return np.mean(x, axis=-1)
    >>>
    >>> def std_feature(x):
    ...     return np.std(x, axis=-1)
    >>>
    >>> # 使用函数列表
    >>> features, names = make_tabular(signals, [mean_feature, std_feature], return_feat_names=True)
    >>> print(features.shape)  # 输出: (10, 8)
    >>> print(names)  # 输出: ['mean_feature_ch0', 'mean_feature_ch1', ..., 'std_feature_ch3']
    >>>
    >>> # 使用字典
    >>> feature_dict = {'Mean': mean_feature, 'Std': std_feature}
    >>> features, names = make_tabular(signals, feature_dict, return_feat_names=True)
    >>> print(features.shape)  # 输出: (10, 8)
    >>> print(names)  # 输出: ['Mean_ch0', 'Mean_ch1', ..., 'Std_ch3']
    """
    signals = ensure_npy(signals)
    signals = ensure_3dims(signals, newaxis='batch')

    n_batch, n_channels, n_times = signals.shape

    # 特征名称, 如果为字典，特征名称就为 key name, 否则为 function 的名称
    if isinstance(fn_features, dict):
        feature_list = list(fn_features.items())
    else:
        feature_list = [(func.__name__, func) for func in fn_features]

    # 使用 fn_features 计算特征
    features = []
    feature_names = []

    for name, func in feature_list:
        feat = func(signals)

        if isinstance(feat, tuple):
            raise ValueError(f'{func} 返回多个值, 不符合要求!')

        scalar = ensure_scalar(feat)
        if scalar is not None:
            # fn 返回标量，则直接加入特征向量中
            features.append(np.full((n_batch, 1), scalar))
            feature_names.append(name)
            continue

        feat = np.squeeze(feat)

        if feat.ndim == 1:
            # fn 返回向量 [N,]，则将向量展平后加入特征向量中
            assert feat.shape[0] == n_batch
            features.append(feat.reshape(-1, 1))
            feature_names.append(name)
        elif feat.ndim == 2:
            # fn 返回 [N, K] 或者 [N, C] 的二维数组，则将特征拼接起来
            # feature name: name{i...k}_ch{j...c}
            assert feat.shape[0] == n_batch

            if feat.shape[1] == n_channels:
                features.append(feat)
                feature_names.extend([f'{name}_ch{j}' for j in range(n_channels)])
            else:
                features.append(feat)
                feature_names.extend([f'{name}{i}' for i in range(feat.shape[1])])
        elif feat.ndim == 3:
            # fn 返回 [N, C, K] 的三维数组，则将特征拼接起来
            # feature name: name{i...k}_ch{j...c}
            assert feat.shape[0] == n_batch and feat.shape[1] == n_channels
            feat_reshaped = feat.reshape(n_batch, -1)
            features.append(feat_reshaped)
            feature_names.extend(
                [
                    f'{name}{i}_ch{j}'
                    for j in range(n_channels)
                    for i in range(feat.shape[2])
                ]
            )
        else:
            raise ValueError(f'Unsupported feature shape: {feat.shape}')

    # 最后全部堆叠起来
    features = np.hstack(features)

    # 输出 (n_batch, n_features)
    if return_feat_names:
        return features, feature_names
    else:
        return features



from collections import Counter
from typing import Optional, Sequence, TypedDict, Union

import numpy as np
import torch
from sklearn.model_selection import KFold
from torch.utils.data import DataLoader, Dataset, Subset
from torch.utils.data import random_split as torch_random_split

from zyplib.utils.ensure import ensure_tensor


class BasicXyDataset(Dataset):
    def __init__(
        self,
        data: Union[np.ndarray, torch.Tensor],
        labels: Union[np.ndarray, torch.Tensor],
    ):
        self.data = ensure_tensor(data, dtype=torch.float32)
        self.labels = ensure_tensor(labels, dtype=torch.float32)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.data[idx], self.labels[idx]


def make_basic_dataloader(
    xy_train: tuple[np.ndarray, np.ndarray],
    xy_test: Optional[tuple[np.ndarray, np.ndarray]] = None,
    split_val_from_train: Sequence[Union[int, float]] = None,
    batch_size: int = 64,
    shuffle: bool = True,
    num_workers: int = 0,
    pin_memory: bool = False,
    **kwargs,
):
    """从训练和测试数据创建 DataLoader(s)。

    参数
    ----------
    - `xy_train` : `tuple[np.ndarray, np.ndarray]`
        - 包含训练数据和标签的元组。
    - `xy_test` : `tuple[np.ndarray, np.ndarray]`, 可选
        - 包含测试数据和标签的元组，默认为 None。
    - `split_val_from_train` : `tuple[float, float] | tuple[int, int]`, 可选
        - 指定如何将训练数据拆分为训练集和验证集，默认为 None。
        - 如果为 None，则不进行拆分。
    - `batch_size` : `int`, 可选
        - 每个批次的样本数量，默认为 64。
    - `shuffle` : `bool`, 可选
        - 是否在每个 epoch 时打乱数据，默认为 True。
        - 只对训练数据有效；val/test 数据集不进行打乱。
    - `num_workers` : `int`, 可选
        - 用于数据加载的子进程数量，默认为 0。
    - `pin_memory` : `bool`, 可选
        - 如果为 True，数据加载器将把张量复制到 CUDA 固定内存中，默认为 False。
    - `**kwargs` : `dict`, 可选
        - DataLoader 的其他参数，默认为 None。

    返回
    ----------
    - `tuple[DataLoader]`
        - 训练、验证和测试数据集的 DataLoader 元组。
    """
    train_dataset = BasicXyDataset(*xy_train)

    if xy_test is not None:
        test_dataset = BasicXyDataset(*xy_test)
    else:
        test_dataset = None

    # 划分验证集
    if split_val_from_train:
        train_dataset, val_dataset = torch_random_split(
            train_dataset, split_val_from_train
        )
    else:
        val_dataset = None

    def make_dataloader(dataset: Dataset, shuffle: bool):
        if dataset is None:
            return None
        return DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=pin_memory,
            **kwargs,
        )

    train_loader = make_dataloader(train_dataset, shuffle=shuffle)
    val_loader = make_dataloader(val_dataset, shuffle=False)
    test_loader = make_dataloader(test_dataset, shuffle=False)

    return tuple(filter(lambda x: x is not None, [train_loader, val_loader, test_loader]))


# alias
make_dataloader_from_xy = make_basic_dataloader


def make_kfold_dataloader(
    *,
    xy: tuple[np.ndarray, np.ndarray] | None = None,
    dataset: Dataset | None = None,
    k: int = 5,
    random_seed: int | None = None,
    batch_size: int = 64,
    num_workers: int = 0,
    pin_memory: bool = False,
    **kwargs,
):
    """根据输入的 xy 或者 dataset 创建 KFold 的 DataLoader(s)。

    > 注意，这里是导入 XY 之后再进行 KFold；一般而言，应该是导入文件的时候就直接进行划分，推荐使用 `make_xy_pipeline` 的 `kfold_pipeline` 函数

    从输入的数据中创建 KFold 的 DataLoader(s)。

    xy 和 dataset 二选一。

    Parameters
    ----------
    - `xy` : `tuple[np.ndarray, np.ndarray]`
        - 包含数据和标签的元组。
    - `dataset` : `Dataset`, 可选
        - 数据集。
    - `k` : `int`, 可选
        - 折数，默认为 5。
    - `random_seed` : `int | None`, 可选
        - 随机种子，默认为 None。
    - `batch_size` : `int`, 可选
        - 每个批次的样本数量，默认为 64。
    - `num_workers` : `int`, 可选
        - 用于数据加载的子进程数量，默认为 0。
    - `pin_memory` : `bool`, 可选
        - 如果为 True，数据加载器将把张量复制到 CUDA 固定内存中，默认为 False。
    - `**kwargs` : `dict`, 可选
        - DataLoader 的其他参数，默认为 None。

    Yields
    ----------
    - `tuple[DataLoader, DataLoader]`
        - 训练和验证数据集的 DataLoader。
    """
    # xy 和 dataset 只能有一个是 None，一个不是 None
    assert (
        xy is not None or dataset is not None
    ), 'xy 和 dataset 只能有一个是 None，一个不是 None'
    if xy is not None:
        dataset = BasicXyDataset(*xy)
    else:
        dataset = dataset

    kf = KFold(n_splits=k, shuffle=True, random_state=random_seed)

    for train_idx, val_idx in kf.split(dataset):
        train_dataset = Subset(dataset, train_idx)
        val_dataset = Subset(dataset, val_idx)
        train_loader = DataLoader(
            train_dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            **kwargs,
        )
        val_loader = DataLoader(
            val_dataset,
            batch_size=batch_size,
            num_workers=num_workers,
            pin_memory=pin_memory,
            **kwargs,
        )
        yield train_loader, val_loader


class LabelCounter(TypedDict):
    number: int
    ratio: float


def count_dataloader_labels(loader: DataLoader) -> dict[int, LabelCounter]:
    """检查数据加载器是否平衡"""
    all_labels = []
    for _, labels in loader:
        labels = labels.numpy()
        all_labels.extend(np.squeeze(labels).tolist())

    counter = Counter(all_labels)
    total_samples = sum(counter.values())
    return {
        label: {'number': number, 'ratio': number / total_samples}
        for label, number in counter.items()
    }
